"""
This module contains the base objects for use in base xml/serializable functionality.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import re
from xml.etree import ElementTree
from weakref import WeakKeyDictionary

import numpy
from numpy.linalg import norm

from sarpy.io.xml.base import DEFAULT_STRICT, get_node_value, find_children, \
    Arrayable, ParametersCollection, SerializableArray, \
    parse_str, parse_bool, parse_int, parse_float, parse_complex, parse_datetime, \
    parse_serializable, parse_serializable_list

logger = logging.getLogger(__name__)

_length_text = 'Specified minimum length is `{}`,\n\t' \
               'while specified maximum length is `{}`'
_type_text = 'Field {} of class {} got incompatible type {}.'


class BasicDescriptor(object):
    """A descriptor object for reusable properties. Note that it is required that the calling instance is hashable."""
    _typ_string = None

    def __init__(self, name, required, strict=DEFAULT_STRICT, default_value=None, docstring=''):
        self.data = WeakKeyDictionary()  # our instance reference dictionary
        # WeakDictionary use is subtle here. A reference to a particular class instance in this dictionary
        # should not be the thing keeping a particular class instance from being destroyed.
        self.name = name
        self.required = (name in required)
        self.strict = strict
        self.default_value = default_value

        self.__doc__ = docstring
        self._format_docstring()

    def _format_docstring(self):
        docstring = self.__doc__
        if docstring is None:
            docstring = ''
        if (self._typ_string is not None) and (not docstring.startswith(self._typ_string)):
            docstring = '{} {}'.format(self._typ_string, docstring)

        suff = self._docstring_suffix()
        if suff is not None:
            docstring = '{} {}'.format(docstring, suff)

        lenstr = self._len_string()
        if lenstr is not None:
            docstring = '{} {}'.format(docstring, lenstr)

        if self.required:
            docstring = '{} {}'.format(docstring, ' **Required.**')
        else:
            docstring = '{} {}'.format(docstring, ' **Optional.**')
        self.__doc__ = docstring

    def _len_string(self):
        minl = getattr(self, 'minimum_length', None)
        maxl = getattr(self, 'minimum_length', None)
        def_minl = getattr(self, '_DEFAULT_MIN_LENGTH', None)
        def_maxl = getattr(self, '_DEFAULT_MAX_LENGTH', None)
        if minl is not None and maxl is not None:
            if minl == def_minl and maxl == def_maxl:
                return None

            lenstr = ' Must have length '
            if minl == def_minl:
                lenstr += '<= {0:d}.'.format(maxl)
            elif maxl == def_maxl:
                lenstr += '>= {0:d}.'.format(minl)
            elif minl == maxl:
                lenstr += ' exactly {0:d}.'.format(minl)
            else:
                lenstr += 'in the range [{0:d}, {1:d}].'.format(minl, maxl)
            return lenstr
        else:
            return None

    def _docstring_suffix(self):
        return None

    def __get__(self, instance, owner):
        """The getter.

        Parameters
        ----------
        instance : object
            the calling class instance
        owner : object
            the type of the class - that is, the actual object to which this descriptor is assigned

        Returns
        -------
        object
            the return value
        """

        if instance is None:
            # this has been access on the class, so return the class
            return self

        fetched = self.data.get(instance, self.default_value)
        if fetched is not None or not self.required:
            return fetched
        else:
            msg = 'Required field {} of class {} is not populated.'.format(self.name, instance.__class__.__name__)
            if self.strict:
                raise AttributeError(msg)
            else:
                logger.debug(msg)  # NB: this is at debug level to not be too verbose
            return fetched

    def __set__(self, instance, value):
        """The setter method.

        Parameters
        ----------
        instance : object
            the calling class instance
        value
            the value to use in setting - the type depends on the specific extension of this base class

        Returns
        -------
        bool
            this base class, and only this base class, handles the required compliance and None behavior and has
            a return. This returns True if this the setting value was None, and False otherwise.
        """

        # NOTE: This is intended to handle this case for every extension of this class. Hence, the boolean return,
        # which extensions SHOULD NOT implement. This is merely to follow DRY principles.
        if value is None:
            if self.default_value is not None:
                self.data[instance] = self.default_value
                return True
            elif self.required:
                if self.strict:
                    raise ValueError(
                        'Attribute {} of class {} cannot be assigned '
                        'None.'.format(self.name, instance.__class__.__name__))
                else:
                    logger.debug(  # NB: this is at debuglevel to not be too verbose
                        'Required attribute {} of class {} has been set to None.'.format(
                            self.name, instance.__class__.__name__))
            self.data[instance] = None
            return True
        # note that the remainder must be implemented in each extension
        return False  # this is probably a bad habit, but this returns something for convenience alone


class StringDescriptor(BasicDescriptor):
    """A descriptor for string type"""
    _typ_string = 'str:'

    def __init__(self, name, required, strict=DEFAULT_STRICT, default_value=None, docstring=None):
        super(StringDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)

    def __set__(self, instance, value):

        if super(StringDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        self.data[instance] = parse_str(value, self.name, instance)


class StringListDescriptor(BasicDescriptor):
    """A descriptor for properties for an array type item for specified extension of string"""
    _typ_string = 'List[str]:'
    _DEFAULT_MIN_LENGTH = 0
    _DEFAULT_MAX_LENGTH = 2 ** 32

    def __init__(self, name, required, strict=DEFAULT_STRICT, minimum_length=None, maximum_length=None,
                 default_value=None, docstring=None):
        self.minimum_length = self._DEFAULT_MIN_LENGTH if minimum_length is None else int(minimum_length)
        self.maximum_length = self._DEFAULT_MAX_LENGTH if maximum_length is None else int(maximum_length)
        if self.minimum_length > self.maximum_length:
            raise ValueError(_length_text.format(self.minimum_length, self.maximum_length))
        super(StringListDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)

    def __set__(self, instance, value):
        def set_value(new_value):
            if len(new_value) < self.minimum_length:
                msg = 'Attribute {} of class {} is a string list of size {},\n\t' \
                      'and must have length at least {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.minimum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.error(msg)
            if len(new_value) > self.maximum_length:
                msg = 'Attribute {} of class {} is a string list of size {},\n\t' \
                      'and must have length no greater than {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.maximum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.error(msg)
            self.data[instance] = new_value

        if super(StringListDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, str):
            set_value([value, ])
        elif isinstance(value, ElementTree.Element):
            set_value([get_node_value(value), ])
        elif isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], str):
                set_value(value)
            elif isinstance(value[0], ElementTree.Element):
                set_value([get_node_value(nod) for nod in value])
        else:
            raise TypeError(_type_text.format(self.name, instance.__class__.__name__, type(value)))


class StringEnumDescriptor(BasicDescriptor):
    """A descriptor for enumerated (specified) string type.
    **This implicitly assumes that the valid entries are upper case.**"""
    _typ_string = 'str:'

    def __init__(self, name, values, required, strict=DEFAULT_STRICT, default_value=None, docstring=None):
        self.values = values
        super(StringEnumDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)
        if (self.default_value is not None) and (self.default_value not in self.values):
            self.default_value = None

    def _docstring_suffix(self):
        suff = ' Takes values in :code:`{}`.'.format(self.values)
        if self.default_value is not None:
            suff += ' Default value is :code:`{}`.'.format(self.default_value)
        return suff

    def __set__(self, instance, value):
        if value is None:
            if self.default_value is not None:
                self.data[instance] = self.default_value
            else:
                super(StringEnumDescriptor, self).__set__(instance, value)
            return

        val = parse_str(value, self.name, instance)

        if val in self.values:
            self.data[instance] = val
        else:
            msg = 'Attribute {} of class {} received {},\n\t' \
                  'but values ARE REQUIRED to be one of {}'.format(
                    self.name, instance.__class__.__name__, value, self.values)
            if self.strict:
                raise ValueError(msg)
            else:
                logger.error(msg)
            self.data[instance] = val


class StringRegexDescriptor(BasicDescriptor):
    """A descriptor for a string matching a regex."""
    _typ_string = 'str:'

    def __init__(self, name, pattern, required, strict=DEFAULT_STRICT, default_value=None, docstring=None):
        self.pattern = pattern
        self.matcher = re.compile(pattern)
        super(StringRegexDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)
        if (self.default_value is not None) and (not self.matcher.fullmatch(self.default_value)):
            self.default_value = None

    def _docstring_suffix(self):
        suff = ' Takes values matching :code:`{}`.'.format(self.pattern)
        if self.default_value is not None:
            suff += ' Default value is :code:`{}`.'.format(self.default_value)
        return suff

    def __set__(self, instance, value):
        if value is None:
            if self.default_value is not None:
                self.data[instance] = self.default_value
            else:
                super(StringRegexDescriptor, self).__set__(instance, value)
            return

        val = parse_str(value, self.name, instance)

        if self.matcher.fullmatch(val):
            self.data[instance] = val
        else:
            msg = 'Attribute {} of class {} received {},\n\t' \
                  'but values ARE REQUIRED to match {}'.format(
                    self.name, instance.__class__.__name__, value, self.pattern)
            if self.strict:
                raise ValueError(msg)
            else:
                logger.error(msg)
            self.data[instance] = val


class BooleanDescriptor(BasicDescriptor):
    """A descriptor for boolean type"""
    _typ_string = 'bool:'

    def __init__(self, name, required, strict=DEFAULT_STRICT, default_value=None, docstring=None):
        super(BooleanDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)

    def __set__(self, instance, value):
        if super(BooleanDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return
        try:
            self.data[instance] = parse_bool(value, self.name, instance)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to `bool`\n\t'
                'for field {} of class {} with exception {} - {}\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None


class IntegerDescriptor(BasicDescriptor):
    """A descriptor for integer type"""
    _typ_string = 'int:'

    def __init__(self, name, required, strict=DEFAULT_STRICT, bounds=None, default_value=None, docstring=None):
        self.bounds = bounds
        super(IntegerDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)
        if (self.default_value is not None) and not self._in_bounds(self.default_value):
            self.default_value = None

    def _docstring_suffix(self):
        if self.bounds is not None:
            return 'Must be in the range [{}, {}]'.format(*self.bounds)
        return ''

    def _in_bounds(self, value):
        if self.bounds is None:
            return True
        return (self.bounds[0] is None or self.bounds[0] <= value) and \
            (self.bounds[1] is None or value <= self.bounds[1])

    def __set__(self, instance, value):
        if super(IntegerDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            iv = parse_int(value, self.name, instance)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to `int`\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None
            return

        if self._in_bounds(iv):
            self.data[instance] = iv
        else:
            msg = 'Attribute {} of class {} is required by standard\n\t' \
                  'to take value between {}. Invalid value {}'.format(
                    self.name, instance.__class__.__name__, self.bounds, iv)
            if self.strict:
                raise ValueError(msg)
            else:
                logger.error(msg)
            self.data[instance] = iv


class IntegerEnumDescriptor(BasicDescriptor):
    """A descriptor for enumerated (specified) integer type"""
    _typ_string = 'int:'

    def __init__(self, name, values, required, strict=DEFAULT_STRICT, default_value=None, docstring=None):
        self.values = values
        super(IntegerEnumDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)
        if (self.default_value is not None) and (self.default_value not in self.values):
            self.default_value = None

    def _docstring_suffix(self):
        return 'Must take one of the values in {}.'.format(self.values)

    def __set__(self, instance, value):
        if super(IntegerEnumDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            iv = parse_int(value, self.name, instance)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to `int`\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None
            return

        if iv in self.values:
            self.data[instance] = iv
        else:
            msg = 'Attribute {} of class {} must take value in {}.\n\t' \
                  'Invalid value {}.'.format(
                    self.name, instance.__class__.__name__, self.values, iv)
            if self.strict:
                raise ValueError(msg)
            else:
                logger.error(msg)
            self.data[instance] = iv


class IntegerListDescriptor(BasicDescriptor):
    """A descriptor for integer list type properties"""
    _typ_string = 'list[int]:'
    _DEFAULT_MIN_LENGTH = 0
    _DEFAULT_MAX_LENGTH = 2 ** 32

    def __init__(self, name, tag_dict, required, strict=DEFAULT_STRICT,
                 minimum_length=None, maximum_length=None, docstring=None):
        self.child_tag = tag_dict[name]['child_tag']
        self.minimum_length = self._DEFAULT_MIN_LENGTH if minimum_length is None else int(minimum_length)
        self.maximum_length = self._DEFAULT_MAX_LENGTH if maximum_length is None else int(maximum_length)
        if self.minimum_length > self.maximum_length:
            raise ValueError(_length_text.format(self.minimum_length, self.maximum_length))
        super(IntegerListDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        def set_value(new_value):
            if len(new_value) < self.minimum_length:
                msg = 'Attribute {} of class {} is an integer list of size {},\n\t' \
                      'and must have size at least {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.minimum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.info(msg)
            if len(new_value) > self.maximum_length:
                msg = 'Attribute {} of class {} is an integer list of size {},\n\t' \
                      'and must have size no larger than {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.maximum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.info(msg)
            self.data[instance] = new_value

        if super(IntegerListDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, int):
            set_value([value, ])
        elif isinstance(value, ElementTree.Element):
            set_value([int(get_node_value(value)), ])
        elif isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], int):
                set_value(value)
            elif isinstance(value[0], ElementTree.Element):
                set_value([int(get_node_value(nod)) for nod in value])
        else:
            raise TypeError(_type_text.format(self.name, instance.__class__.__name__, type(value)))


class FloatDescriptor(BasicDescriptor):
    """A descriptor for float type properties"""
    _typ_string = 'float:'

    def __init__(self, name, required, strict=DEFAULT_STRICT, bounds=None, default_value=None, docstring=None):
        self.bounds = bounds
        super(FloatDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)
        if (self.default_value is not None) and not self._in_bounds(self.default_value):
            self.default_value = None

    def _docstring_suffix(self):
        if self.bounds is not None:
            return 'Must be in the range [{}, {}]'.format(*self.bounds)
        return ''

    def _in_bounds(self, value):
        if self.bounds is None:
            return True

        return (self.bounds[0] is None or self.bounds[0] <= value) and \
            (self.bounds[1] is None or value <= self.bounds[1])

    def __set__(self, instance, value):
        if super(FloatDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            iv = parse_float(value, self.name, instance)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to `float`\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None
            return

        if self._in_bounds(iv):
            self.data[instance] = iv
        else:
            msg = 'Attribute {} of class {}\n\t' \
                  'is required by standard to take value between {}.'.format(
                    self.name, instance.__class__.__name__, self.bounds)
            if self.strict:
                raise ValueError(msg)
            else:
                logger.info(msg)
            self.data[instance] = iv


class FloatListDescriptor(BasicDescriptor):
    """A descriptor for float list type properties"""
    _typ_string = 'list[float]:'
    _DEFAULT_MIN_LENGTH = 0
    _DEFAULT_MAX_LENGTH = 2 ** 32

    def __init__(self, name, tag_dict, required, strict=DEFAULT_STRICT,
                 minimum_length=None, maximum_length=None, docstring=None):
        self.child_tag = tag_dict[name]['child_tag']
        self.minimum_length = self._DEFAULT_MIN_LENGTH if minimum_length is None else int(minimum_length)
        self.maximum_length = self._DEFAULT_MAX_LENGTH if maximum_length is None else int(maximum_length)
        if self.minimum_length > self.maximum_length:
            raise ValueError(_length_text.format(self.minimum_length, self.maximum_length))
        super(FloatListDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        def set_value(new_value):
            if len(new_value) < self.minimum_length:
                msg = 'Attribute {} of class {} is an float list of size {},\n\t' \
                      'and must have size at least {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.minimum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.info(msg)
            if len(new_value) > self.maximum_length:
                msg = 'Attribute {} of class {} is a float list of size {},\n\t' \
                      'and must have size no larger than {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.maximum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.info(msg)
            self.data[instance] = new_value

        if super(FloatListDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, float):
            set_value([value, ])
        elif isinstance(value, ElementTree.Element):
            set_value([float(get_node_value(value)), ])
        elif isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], float):
                set_value(value)
            elif isinstance(value[0], ElementTree.Element):
                set_value([float(get_node_value(nod)) for nod in value])
        else:
            raise TypeError(_type_text.format(self.name, instance.__class__.__name__, type(value)))


class ComplexDescriptor(BasicDescriptor):
    """A descriptor for complex valued properties"""
    _typ_string = 'complex:'

    def __init__(self, name, required, strict=DEFAULT_STRICT, default_value=None, docstring=None):
        super(ComplexDescriptor, self).__init__(
            name, required, strict=strict, default_value=default_value, docstring=docstring)

    def __set__(self, instance, value):
        if super(ComplexDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            self.data[instance] = parse_complex(value, self.name, instance)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to `complex`\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None


class FloatArrayDescriptor(BasicDescriptor):
    """A descriptor for float array type properties"""
    _DEFAULT_MIN_LENGTH = 0
    _DEFAULT_MAX_LENGTH = 2 ** 32
    _typ_string = 'numpy.ndarray[float64]:'

    def __init__(self, name, tag_dict, required, strict=DEFAULT_STRICT, minimum_length=None, maximum_length=None,
                 docstring=None):

        self.child_tag = tag_dict[name]['child_tag']
        self.minimum_length = self._DEFAULT_MIN_LENGTH if minimum_length is None else int(minimum_length)
        self.maximum_length = self._DEFAULT_MAX_LENGTH if maximum_length is None else int(maximum_length)
        if self.minimum_length > self.maximum_length:
            raise ValueError(_length_text.format(self.minimum_length, self.maximum_length))
        super(FloatArrayDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        def set_value(new_val):
            if len(new_val) < self.minimum_length:
                msg = 'Attribute {} of class {} is a double array of size {},\n\t' \
                      'and must have size at least {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.minimum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.error(msg)
            if len(new_val) > self.maximum_length:
                msg = 'Attribute {} of class {} is a double array of size {},\n\t' \
                      'and must have size no larger than {}.'.format(
                        self.name, instance.__class__.__name__, value.size, self.maximum_length)
                if self.strict:
                    raise ValueError(msg)
                else:
                    logger.error(msg)
            self.data[instance] = new_val

        if super(FloatArrayDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, numpy.ndarray):
            if not (len(value) == 1) and (numpy.dtype.name == 'float64'):
                raise ValueError('Only one-dimensional ndarrays of dtype float64 are supported here.')
            set_value(value)
        elif isinstance(value, ElementTree.Element):
            xml_ns = getattr(instance, '_xml_ns', None)
            # noinspection PyProtectedMember
            if hasattr(instance, '_child_xml_ns_key') and self.name in instance._child_xml_ns_key:
                # noinspection PyProtectedMember
                xml_ns_key = instance._child_xml_ns_key[self.name]
            else:
                xml_ns_key = getattr(instance, '_xml_ns_key', None)

            size = int(value.attrib['size'])
            child_nodes = find_children(value, self.child_tag, xml_ns, xml_ns_key)
            if len(child_nodes) != size:
                raise ValueError(
                    'Field {} of double array type functionality belonging to class {} got a ElementTree element '
                    'with size attribute {}, but has {} child nodes with tag {}.'.format(
                        self.name, instance.__class__.__name__, size, len(child_nodes), self.child_tag))
            new_value = numpy.empty((size,), dtype=numpy.float64)
            for i, node in enumerate(child_nodes):
                new_value[i] = float(get_node_value(node))
            set_value(new_value)
        elif isinstance(value, list):
            # user or json deserialization
            set_value(numpy.array(value, dtype=numpy.float64))
        else:
            raise TypeError(_type_text.format(self.name, instance.__class__.__name__, type(value)))


class DateTimeDescriptor(BasicDescriptor):
    """A descriptor for date time type properties"""
    _typ_string = 'numpy.datetime64:'

    def __init__(self, name, required, strict=DEFAULT_STRICT, docstring=None, numpy_datetime_units='us'):
        self.units = numpy_datetime_units  # s, ms, us, ns are likely choices here, depending on needs
        super(DateTimeDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(DateTimeDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return
        self.data[instance] = parse_datetime(value, self.name, instance, self.units)


class FloatModularDescriptor(BasicDescriptor):
    """
    A descriptor for float type which will take values in a range [-limit, limit], set using modular
    arithmetic operations
    """
    _typ_string = 'float:'

    def __init__(self, name, limit, required, strict=DEFAULT_STRICT, docstring=None):
        self.limit = float(limit)
        super(FloatModularDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(FloatModularDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            val = parse_float(value, self.name, instance)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to `float`\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None
            return

        # do modular arithmetic manipulations
        val = (val % (2 * self.limit))  # NB: % and * have same precedence, so it can be super dumb
        self.data[instance] = val if val <= self.limit else val - 2 * self.limit


class SerializableDescriptor(BasicDescriptor):
    """A descriptor for properties of a specified type assumed to be an extension of Serializable"""

    def __init__(self, name, the_type, required, strict=DEFAULT_STRICT, docstring=None):
        self.the_type = the_type
        self._typ_string = str(the_type).strip().split('.')[-1][:-2] + ':'
        super(SerializableDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(SerializableDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            self.data[instance] = parse_serializable(value, self.name, instance, self.the_type)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to Serializable type {}\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard.'.format(
                    value, type(value), self.the_type, self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None


class UnitVectorDescriptor(BasicDescriptor):
    """A descriptor for properties of a specified type assumed to be of subtype of Arrayable"""

    def __init__(self, name, the_type, required, strict=DEFAULT_STRICT, docstring=None):
        if not issubclass(the_type, Arrayable):
            raise TypeError(
                'The input type {} for field {} must be a subclass of Arrayable.'.format(the_type, name))
        self.the_type = the_type

        self._typ_string = str(the_type).strip().split('.')[-1][:-2] + ':'
        super(UnitVectorDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(UnitVectorDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            vec = parse_serializable(value, self.name, instance, self.the_type)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to Unit Vector Type type {}\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.the_type, self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None
            return None

        # noinspection PyTypeChecker
        coords = vec.get_array(dtype=numpy.float64)
        the_norm = norm(coords)
        if the_norm == 0:
            logger.error(
                'The input for field {} is expected to be made into a unit vector.\n\t'
                'In this case, the norm of the input is 0.\n\t'
                'The value is set to None, which may be against the standard.'.format(
                    self.name))
            self.data[instance] = None
        elif the_norm == 1:
            self.data[instance] = vec
        else:
            self.data[instance] = self.the_type.from_array(coords/the_norm)


class ParametersDescriptor(BasicDescriptor):
    """A descriptor for properties of a Parameter type - that is, dictionary"""

    def __init__(self, name, tag_dict, required, strict=DEFAULT_STRICT, docstring=None):
        self.child_tag = tag_dict[name]['child_tag']
        self._typ_string = 'ParametersCollection:'
        super(ParametersDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(ParametersDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, ParametersCollection):
            self.data[instance] = value
        else:
            the_inst = self.data.get(instance, None)
            xml_ns = getattr(instance, '_xml_ns', None)
            # noinspection PyProtectedMember
            if hasattr(instance, '_child_xml_ns_key') and self.name in instance._child_xml_ns_key:
                # noinspection PyProtectedMember
                xml_ns_key = instance._child_xml_ns_key[self.name]
            else:
                xml_ns_key = getattr(instance, '_xml_ns_key', None)
            if the_inst is None:
                self.data[instance] = ParametersCollection(
                    collection=value, name=self.name, child_tag=self.child_tag,
                    _xml_ns=xml_ns, _xml_ns_key=xml_ns_key)
            else:
                the_inst.set_collection(value)


class SerializableListDescriptor(BasicDescriptor):
    """A descriptor for properties of a list or array of specified extension of Serializable"""

    def __init__(self, name, child_type, tag_dict, required, strict=DEFAULT_STRICT, docstring=None):
        self.child_type = child_type
        tags = tag_dict[name]
        self.array = tags.get('array', False)
        if self.array:
            raise ValueError(
                'Attribute {} is populated in the `_collection_tags` dictionary with `array`=True. '
                'This is inconsistent with using SerializableListDescriptor.'.format(name))

        self.child_tag = tags['child_tag']
        self._typ_string = 'List[{}]:'.format(str(child_type).strip().split('.')[-1][:-2])
        super(SerializableListDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(SerializableListDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        try:
            self.data[instance] = parse_serializable_list(value, self.name, instance, self.child_type)
        except Exception as e:
            logger.error(
                'Failed converting {} of type {} to serializable list of type {}\n\t'
                'for field {} of class {} with exception {} - {}.\n\t'
                'Setting value to None, which may be against the standard'.format(
                    value, type(value), self.child_type, self.name, instance.__class__.__name__, type(e), e))
            self.data[instance] = None


class SerializableArrayDescriptor(BasicDescriptor):
    """A descriptor for properties of a list or array of specified extension of Serializable"""
    _DEFAULT_MIN_LENGTH = 0
    _DEFAULT_MAX_LENGTH = 2 ** 32

    def __init__(self, name, child_type, tag_dict, required, strict=DEFAULT_STRICT,
                 minimum_length=None, maximum_length=None, docstring=None,
                 array_extension=SerializableArray):
        if not issubclass(array_extension, SerializableArray):
            raise TypeError('array_extension must be a subclass of SerializableArray.')
        self.child_type = child_type
        tags = tag_dict[name]
        self.array = tags.get('array', False)
        if not self.array:
            raise ValueError(
                'Attribute {} is populated in the `_collection_tags` dictionary without `array`=True. '
                'This is inconsistent with using SerializableArrayDescriptor.'.format(name))

        self.child_tag = tags['child_tag']
        self._typ_string = 'numpy.ndarray[{}]:'.format(str(child_type).strip().split('.')[-1][:-2])
        self.array_extension = array_extension

        self.minimum_length = self._DEFAULT_MIN_LENGTH if minimum_length is None else int(minimum_length)
        self.maximum_length = self._DEFAULT_MAX_LENGTH if maximum_length is None else int(maximum_length)
        if self.minimum_length > self.maximum_length:
            raise ValueError(_length_text.format(self.minimum_length, self.maximum_length))
        super(SerializableArrayDescriptor, self).__init__(name, required, strict=strict, docstring=docstring)

    def __set__(self, instance, value):
        if super(SerializableArrayDescriptor, self).__set__(instance, value):  # the None handler...kinda hacky
            return

        if isinstance(value, self.array_extension):
            self.data[instance] = value
        else:
            xml_ns = getattr(instance, '_xml_ns', None)
            xml_ns_key = getattr(instance, '_xml_ns_key', None)
            the_inst = self.data.get(instance, None)
            if the_inst is None:
                self.data[instance] = self.array_extension(
                    coords=value, name=self.name, child_tag=self.child_tag, child_type=self.child_type,
                    minimum_length=self.minimum_length, maximum_length=self.maximum_length, _xml_ns=xml_ns,
                    _xml_ns_key=xml_ns_key)
            else:
                the_inst.set_array(value)




from ..tre_elements import TREExtension, TREElement

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class ICHIPBType(TREElement):
    def __init__(self, value):
        super(ICHIPBType, self).__init__()
        self.add_field('XFRM_FLAG', 's', 2, value)
        self.add_field('SCALE_FACTOR', 's', 10, value)
        self.add_field('ANAMRPH_CORR', 's', 2, value)
        self.add_field('SCANBLK_NUM', 's', 2, value)
        self.add_field('OP_ROW_11', 's', 12, value)
        self.add_field('OP_COL_11', 's', 12, value)
        self.add_field('OP_ROW_12', 's', 12, value)
        self.add_field('OP_COL_12', 's', 12, value)
        self.add_field('OP_ROW_21', 's', 12, value)
        self.add_field('OP_COL_21', 's', 12, value)
        self.add_field('OP_ROW_22', 's', 12, value)
        self.add_field('OP_COL_22', 's', 12, value)
        self.add_field('FI_ROW_11', 's', 12, value)
        self.add_field('FI_COL_11', 's', 12, value)
        self.add_field('FI_ROW_12', 's', 12, value)
        self.add_field('FI_COL_12', 's', 12, value)
        self.add_field('FI_ROW_21', 's', 12, value)
        self.add_field('FI_COL_21', 's', 12, value)
        self.add_field('FI_ROW_22', 's', 12, value)
        self.add_field('FI_COL_22', 's', 12, value)
        self.add_field('FI_ROW', 's', 8, value)
        self.add_field('FI_COL', 's', 8, value)


class ICHIPB(TREExtension):
    _tag_value = 'ICHIPB'
    _data_type = ICHIPBType




from ..tre_elements import TREExtension, TREElement
import struct

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


class FRAME(TREElement):
    def __init__(self, value):
        super(FRAME, self).__init__()
        self.add_field('FRAME_OFFSET', 'b', 4, value)


class NBLOCAType(TREElement):
    def __init__(self, value):
        super(NBLOCAType, self).__init__()
        self.add_field('FRAME_1_OFFSET', 'b', 4, value)
        self.add_field('NUMBER_OF_FRAMES', 'b', 4, value)
        self.add_loop('FRAMEs', struct.unpack('>I', self.NUMBER_OF_FRAMES)[0]-1, FRAME, value)


class NBLOCA(TREExtension):
    _tag_value = 'NBLOCA'
    _data_type = NBLOCAType



"""
Functionality for reading TerraSAR-X data into a SICD model.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import os
import logging
from xml.etree import ElementTree
from typing import Union, List, Tuple, Optional, BinaryIO
from functools import reduce
import struct

import numpy
import numpy.linalg
from numpy.polynomial import polynomial
from scipy.constants import speed_of_light

from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd_elements.blocks import Poly1DType, Poly2DType
from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd_elements.CollectionInfo import CollectionInfoType, RadarModeType
from sarpy.io.complex.sicd_elements.ImageCreation import ImageCreationType
from sarpy.io.complex.sicd_elements.RadarCollection import RadarCollectionType, \
    WaveformParametersType, ChanParametersType, TxStepType
from sarpy.io.complex.sicd_elements.ImageData import ImageDataType
from sarpy.io.complex.sicd_elements.GeoData import GeoDataType, SCPType
from sarpy.io.complex.sicd_elements.Position import PositionType, XYZPolyType
from sarpy.io.complex.sicd_elements.Grid import GridType, DirParamType, WgtTypeType
from sarpy.io.complex.sicd_elements.Timeline import TimelineType, IPPSetType
from sarpy.io.complex.sicd_elements.ImageFormation import ImageFormationType, \
    RcvChanProcType
from sarpy.io.complex.sicd_elements.RMA import RMAType, INCAType
from sarpy.io.complex.sicd_elements.Radiometric import RadiometricType, NoiseLevelType_
from sarpy.io.complex.utils import two_dim_poly_fit, fit_position_xvalidation

from sarpy.io.general.base import SarpyIOError
from sarpy.io.general.data_segment import DataSegment, NumpyMemmapSegment, SubsetSegment
from sarpy.io.general.format_function import ComplexFormatFunction
from sarpy.io.general.utils import get_seconds, parse_timestring, is_file_like


logger = logging.getLogger(__name__)


##########
# helper functions and basic interpreter

def _parse_xml(file_name: str, without_ns: bool = False) -> Union[
        ElementTree.Element, Tuple[dict, ElementTree.Element]]:
    root_node = ElementTree.parse(file_name).getroot()
    if without_ns:
        return root_node
    else:
        ns = dict([node for _, node in ElementTree.iterparse(file_name, events=('start-ns', ))])
        return ns, root_node


def _is_level1_product(prospective_file: str) -> bool:
    with open(prospective_file, 'rb') as fi:
        check = fi.read(200)
        if check.startswith(b'<?xml'):
            end_xml_declaration = check.find(b'?>')
            if end_xml_declaration == -1:
                raise ValueError('Poorly formed xml declaration\n\t`{}`'.format(check))
            check = check[end_xml_declaration+2:].strip()
        else:
            check = check.strip()
    return check.startswith(b'<level1Product')


############
# metadata helper class

class TSXDetails(object):
    """
    Parser and interpreter for the TerraSAR-X file package meta-data.
    """

    __slots__ = (
        '_parent_directory', '_main_file', '_georef_file', '_main_root', '_georef_root',
        '_im_format')

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
            The top-level directory, or the basic package xml file.
        """

        self._parent_directory = None
        self._main_file = None
        self._georef_file = None
        self._main_root = None
        self._georef_root = None
        self._im_format = None
        self._validate_file(file_name)
        self._im_format = self._find_main('./productInfo/imageDataInfo/imageDataFormat').text
        if self._im_format not in ['COSAR', 'GEOTIFF']:
            raise ValueError(
                'The file is determined to be of type TerraSAR-X, but we got '
                'unexpected image format value {}'.format(self.image_format))

    def _validate_file(self, file_name: str) -> None:
        """
        Validate the input file location.

        Parameters
        ----------
        file_name : str

        Returns
        -------
        None
        """

        if not isinstance(file_name, str):
            raise SarpyIOError('file_name must be of string type.')
        if not os.path.exists(file_name):
            raise SarpyIOError('file {} does not exist'.format(file_name))

        found_file = None
        if os.path.isdir(file_name):
            for entry in os.listdir(file_name):
                prop_file = os.path.join(file_name, entry)
                if os.path.isfile(prop_file) and os.path.splitext(prop_file)[1] == '.xml' \
                        and _is_level1_product(prop_file):
                    found_file = prop_file

            if found_file is None:
                raise SarpyIOError(
                    'The provided argument is a directory, but we found no level1Product xml file at the top level.')
        elif os.path.splitext(file_name)[1] == '.xml':
            if _is_level1_product(file_name):
                found_file = file_name
            else:
                raise SarpyIOError(
                    'The provided argument is an xml file, which is not a level1Product xml file.')
        else:
            raise SarpyIOError(
                'The provided argument is an file, but does not have .xml extension.')

        if file_name is None:
            raise ValueError('Unspecified error where main_file is not defined.')

        parent_directory = os.path.split(found_file)[0]
        if not os.path.isdir(os.path.join(parent_directory, 'IMAGEDATA')):
            raise ValueError(
                'The input file was determined to be or contain a TerraSAR-X level 1 product file, '
                'but the IMAGEDATA directory is not in the expected relative location.')
        self._parent_directory = parent_directory
        self._main_file = found_file
        self._main_root = _parse_xml(self._main_file, without_ns=True)

        georef_file = os.path.join(parent_directory, 'ANNOTATION', 'GEOREF.xml')
        if not os.path.isfile(georef_file):
            logger.warning(
                'The input file was determined to be or contain a TerraSAR-X level 1 product file,\n\t'
                'but the ANNOTATION/GEOREF.xml is not in the expected relative location.')
        else:
            self._georef_file = georef_file
            self._georef_root = _parse_xml(self._georef_file, without_ns=True)

    @property
    def file_name(self) -> str:
        """
        str: the package directory location
        """

        return self._parent_directory

    @property
    def image_format(self) -> str:
        """
        str: The image file format enum value.
        """

        return self._im_format

    def _find_main(self, tag: str) -> ElementTree.Element:
        """
        Pass through to ElementTree.Element.find(tag).

        Parameters
        ----------
        tag : str

        Returns
        -------
        ElementTree.Element
        """

        return self._main_root.find(tag)

    def _findall_main(self, tag: str) -> List[ElementTree.Element]:
        """
        Pass through to ElementTree.Element.findall(tag).

        Parameters
        ----------
        tag : str

        Returns
        -------
        List[ElementTree.Element
        """

        return self._main_root.findall(tag)

    def _find_georef(self, tag: str) -> ElementTree.Element:
        """
        Pass through to ElementTree.Element.find(tag).

        Parameters
        ----------
        tag : str

        Returns
        -------
        ElementTree.Element
        """

        return None if self._georef_root is None else self._georef_root.find(tag)

    def _findall_georef(self, tag: str) -> List[ElementTree.Element]:
        """
        Pass through to ElementTree.Element.findall(tag).

        Parameters
        ----------
        tag : str

        Returns
        -------
        List[ElementTree.Element
        """

        return None if self._georef_root is None else self._georef_root.findall(tag)

    def _get_state_vector_data(self) -> Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
        """
        Gets the state vector data.

        Returns
        -------
        times: numpy.ndarray
        positions: numpy.ndarray
        velocities: numpy.ndarray
        """

        state_vecs = self._findall_main('./platform/orbit/stateVec')
        tims = numpy.zeros((len(state_vecs),), dtype='datetime64[us]')
        pos = numpy.zeros((len(state_vecs), 3), dtype='float64')
        vel = numpy.zeros((len(state_vecs), 3), dtype='float64')
        for i, entry in enumerate(state_vecs):
            tims[i] = parse_timestring(entry.find('./timeUTC').text, precision='us')
            pos[i, :] = [
                float(entry.find('./posX').text), float(entry.find('./posY').text),
                float(entry.find('./posZ').text)]
            vel[i, :] = [
                float(entry.find('./velX').text), float(entry.find('./velY').text),
                float(entry.find('./velZ').text)]
        return tims, pos, vel

    @staticmethod
    def _parse_pol_string(str_in: str) -> Tuple[str, str]:
        return str_in[0], str_in[1]

    def _get_sicd_tx_rcv_pol(self, str_in: str) -> str:
        tx_pol, rcv_pol = self._parse_pol_string(str_in)
        return '{}:{}'.format(tx_pol, rcv_pol)

    def _get_full_pol_list(self) -> Tuple[List[str], List[str], List[str]]:
        """
        Gets the full list of polarization states.

        Returns
        -------
        original_pols : List[str]
        tx_pols : List[str]
        tx_rcv_pols: List[str]
        """

        t_original_pols = []
        t_tx_pols = []
        t_tx_rcv_pols = []
        # TODO: this in particular is probably troubled for ScanSAR mode
        for node in self._findall_main('./productComponents/imageData'):
            orig_pol = node.find('./polLayer').text
            tx_part, rcv_part = self._parse_pol_string(orig_pol)
            t_original_pols.append(orig_pol)
            t_tx_pols.append(tx_part)
            t_tx_rcv_pols.append('{}:{}'.format(tx_part, rcv_part))
        return t_original_pols, t_tx_pols, t_tx_rcv_pols

    def _find_middle_grid_node(self) -> Optional[ElementTree.Element]:
        """
        Find and returns the middle geolocationGrid point, if it exists.
        Otherwise, returns None.

        Returns
        -------
        None|ElementTree.Element
        """

        if self._georef_root is None:
            return None

        # determine the middle grid location
        az_grid_pts = int(self._find_georef('./geolocationGrid/numberOfGridPoints/azimuth').text)
        rg_grid_pts = int(self._find_georef('./geolocationGrid/numberOfGridPoints/range').text)
        mid_az = int(round(az_grid_pts/2.0)) + 1
        mid_rg = int(round(rg_grid_pts/2.0)) + 1
        test_nodes = self._findall_georef('./geolocationGrid/gridPoint[@iaz="{}"]'.format(mid_az))
        for entry in test_nodes:
            if entry.attrib['irg'] == '{}'.format(mid_rg):
                return entry
        return test_nodes[int(len(test_nodes)/2)]

    def _calculate_dop_polys(self,
                             layer_index: str,
                             azimuth_time_scp: float,
                             range_time_scp: float,
                             collect_start: numpy.datetime64,
                             doppler_rate_reference_node: ElementTree.Element) -> Tuple[numpy.ndarray, numpy.ndarray]:
        """
        Calculate the doppler centroid polynomials. This is apparently extracted
        from the paper "TerraSAR-X Deskew Description" by Michael Stewart dated
        December 11, 2008.

        Parameters
        ----------
        layer_index : str
            The layer index string, required for extracting correct metadata.
        azimuth_time_scp : float
            This is in seconds relative to the collection start.
        range_time_scp : float
            This is in seconds.
        collect_start : numpy.datetime64
            The collection start time.
        doppler_rate_reference_node : ElementTree.Element

        Returns
        -------
        (numpy.ndarray, numpy.ndarray)
        """

        # parse the doppler centroid estimates nodes
        doppler_estimate_nodes = self._findall_main(
            './processing/doppler/dopplerCentroid[@layerIndex="{}"]/dopplerEstimate'.format(layer_index))
        # find the center node and extract some reference parameters
        doppler_count = len(doppler_estimate_nodes)
        doppler_estimate_center_node = doppler_estimate_nodes[int(doppler_count/2)]
        rg_ref_time = float(doppler_estimate_center_node.find('./combinedDoppler/referencePoint').text)

        # extract the doppler centroid information from all the nodes
        diff_times_raw = numpy.zeros((doppler_count, ), dtype='float64')  # offsets from reference time, in seconds
        doppler_range_min = numpy.zeros((doppler_count, ), dtype='float64')  # offsets in seconds
        doppler_range_max = numpy.zeros((doppler_count, ), dtype='float64')  # offsets in seconds
        doppler_poly_est = []
        for i, node in enumerate(doppler_estimate_nodes):
            diff_times_raw[i] = get_seconds(
                parse_timestring(node.find('./timeUTC').text, precision='us'),
                collect_start, precision='us')
            combined_node = node.find('./combinedDoppler')
            doppler_range_min[i] = float(combined_node.find('./validityRangeMin').text)
            doppler_range_max[i] = float(combined_node.find('./validityRangeMax').text)
            doppler_poly_est.append(
                numpy.array([float(entry.text) for entry in combined_node.findall('./coefficient')], dtype='float64'))

        # parse the doppler rate estimate from our provided reference node
        fm_dop = float(doppler_rate_reference_node.find('./dopplerRatePolynomial/coefficient[@exponent="0"]').text)
        ss_zd_s = float(self._find_main('./productInfo/imageDataInfo/imageRaster/columnSpacing').text)
        side_of_track = self._find_main('./productInfo/acquisitionInfo/lookDirection').text[0].upper()
        ss_zd_m = float(self._find_main('./productSpecific/complexImageInfo/projectedSpacingAzimuth').text)
        use_ss_zd_s = -ss_zd_s if side_of_track == 'L' else ss_zd_s

        # create a sampled doppler centroid grid
        range_samples = 49  # this is suggested in the paper
        time_coa = numpy.zeros((doppler_count, range_samples), dtype='float64')
        diff_t_range = numpy.zeros((doppler_count, range_samples), dtype='float64')
        dopp_centroid = numpy.zeros((doppler_count, range_samples), dtype='float64')
        for i, entry in enumerate(diff_times_raw):
            time_coa[i, :] = entry
            diff_t_range[i, :] = numpy.linspace(
                doppler_range_min[i], doppler_range_max[i], num=range_samples) - rg_ref_time
            dopp_centroid[i, :] = polynomial.polyval(diff_t_range[i, :], doppler_poly_est[i])
        diff_t_zd = time_coa - dopp_centroid/fm_dop
        coords_rg = 0.5*(diff_t_range + rg_ref_time - range_time_scp)*speed_of_light
        coords_az = ss_zd_m*(diff_t_zd - azimuth_time_scp)/use_ss_zd_s
        # perform our fitting
        poly_order = 3
        dop_centroid_poly, residuals, rank, sing_values = two_dim_poly_fit(
            coords_rg, coords_az, dopp_centroid,
            x_order=poly_order, y_order=poly_order, x_scale=1e-3, y_scale=1e-3, rcond=1e-35)
        logger.info(
            'The dop centroid polynomial fit details:\n\t'
            'root mean square residuals = {}\n\t'
            'rank = {}\n\t'
            'singular values = {}'.format(residuals, rank, sing_values))

        time_coa_poly, residuals, rank, sing_values = two_dim_poly_fit(
            coords_rg, coords_az, time_coa,
            x_order=poly_order, y_order=poly_order, x_scale=1e-3, y_scale=1e-3, rcond=1e-35)
        logger.info(
            'The dop centroid polynomial fit details:\n\t'
            'root mean square residuals = {}\n\t'
            'rank = {}\n\t'
            'singular values = {}'.format(residuals, rank, sing_values))

        return dop_centroid_poly, time_coa_poly

    def _get_basic_sicd_shell(self,
                              center_freq: float,
                              dop_bw: float,
                              ss_zd_s: float) -> SICDType:
        """
        Define the common sicd elements.

        Parameters
        ----------
        center_freq : float
            The center frequency.
        dop_bw : float
            The doppler bandwidth.
        ss_zd_s : float
            The (positive) zero doppler spacing in the time domain.

        Returns
        -------
        SICDType
        """

        def get_collection_info():
            # type: () -> CollectionInfoType
            collector_name = self._find_main('./productInfo/missionInfo/mission').text
            core_name = self._find_main('./productInfo/sceneInfo/sceneID').text

            mode_id = self._find_main('./productInfo/acquisitionInfo/imagingMode').text
            if mode_id == 'ST':
                # TSX "staring" mode, corresponds to SICD spotlight
                mode_type = 'SPOTLIGHT'
            elif mode_id in ['SL', 'HS']:
                # confusing, but TSX mode "spolight" and "high-resolution spotlight",
                # which actually has a moving beam
                mode_type = 'DYNAMIC STRIPMAP'
            elif mode_id == 'SM':
                # TSX stripmap mode
                mode_type = 'STRIPMAP'
            elif mode_id == 'SC':
                # TSX scansar mode
                mode_type = 'STRIPMAP'
            else:
                raise ValueError('Got unexpected mode id {}'.format(mode_id))

            return CollectionInfoType(
                CollectorName=collector_name,
                CoreName=core_name,
                CollectType='MONOSTATIC',
                RadarMode=RadarModeType(ModeID=mode_id, ModeType=mode_type),
                Classification='UNCLASSIFIED')

        def get_image_creation():
            # type: () -> ImageCreationType
            from sarpy.__about__ import __version__

            create_time = self._find_main('./generalHeader/generationTime').text
            site = self._find_main('./productInfo/generationInfo/level1ProcessingFacility').text
            app_node = self._find_main('./generalHeader/generationSystem')
            application = 'Unknown' if app_node is None else \
                '{} {}'.format(app_node.text, app_node.attrib.get('version', 'version_unknown'))
            return ImageCreationType(Application=application,
                                     DateTime=parse_timestring(create_time, precision='us'),
                                     Site=site,
                                     Profile='sarpy {}'.format(__version__))

        def get_initial_grid():
            # type: () -> GridType
            proj_string = self._find_main('./setup/orderInfo/projection').text
            if proj_string == 'GROUNDRANGE':
                image_plane = 'GROUND'
            elif proj_string == 'SLANTRANGE':
                image_plane = 'SLANT'
            else:
                logger.warning('Got image projection {}'.format(proj_string))
                image_plane = 'OTHER'

            the_type = None
            if self._find_main('./productSpecific/complexImageInfo/imageCoordinateType').text == 'ZERODOPPLER':
                the_type = 'RGZERO'

            row_ss = 0.5*float(
                self._find_main('./productInfo/imageDataInfo/imageRaster/rowSpacing').text)*speed_of_light
            row_bw = 2*float(
                self._find_main('./processing/processingParameter/rangeLookBandwidth').text)/speed_of_light
            row_win_name = self._find_main(
                './processing/processingParameter/rangeWindowID').text.upper()
            row_wgt_type = WgtTypeType(WindowName=row_win_name)
            if row_win_name == 'HAMMING':
                row_wgt_type.Parameters = {
                    'COEFFICIENT': self._find_main('./processing/processingParameter/rangeWindowCoefficient').text}
            row = DirParamType(
                SS=row_ss,
                Sgn=-1,
                ImpRespBW=row_bw,
                KCtr=2*center_freq/speed_of_light,
                DeltaK1=-0.5*row_bw,
                DeltaK2=0.5*row_bw,
                DeltaKCOAPoly=[[0, ], ],
                WgtType=row_wgt_type)

            col_ss = float(self._find_main('./productSpecific/complexImageInfo/projectedSpacingAzimuth').text)
            col_win_name = self._find_main('./processing/processingParameter/azimuthWindowID').text.upper()
            col_wgt_type = WgtTypeType(WindowName=col_win_name)
            if col_win_name == 'HAMMING':
                col_wgt_type.Parameters = {
                    'COEFFICIENT': self._find_main('./processing/processingParameter/azimuthWindowCoefficient').text}
            col = DirParamType(
                SS=col_ss,
                Sgn=-1,
                ImpRespBW=dop_bw*ss_zd_s/col_ss,
                KCtr=0,
                WgtType=col_wgt_type)

            return GridType(
                ImagePlane=image_plane,
                Type=the_type,
                Row=row,
                Col=col)

        def get_initial_image_formation():
            # type: () -> ImageFormationType
            return ImageFormationType(
                RcvChanProc=RcvChanProcType(NumChanProc=1, PRFScaleFactor=1),  # ChanIndex set later
                ImageFormAlgo='RMA',
                TStartProc=0,
                TEndProc=0,  # corrected later
                ImageBeamComp='SV',
                AzAutofocus='NO',
                RgAutofocus='NO',
                STBeamComp='SV' if collection_info.RadarMode.ModeID in ['SL', 'HS'] else 'GLOBAL')
            # NB: SL and HS are the proper spotlight modes

        def get_initial_rma():
            # type: () -> RMAType
            return RMAType(RMAlgoType='OMEGA_K',
                           INCA=INCAType(FreqZero=center_freq))

        collection_info = get_collection_info()
        image_creation = get_image_creation()
        init_grid = get_initial_grid()
        init_image_formation = get_initial_image_formation()
        init_rma = get_initial_rma()

        return SICDType(
            CollectionInfo=collection_info,
            ImageCreation=image_creation,
            Grid=init_grid,
            ImageFormation=init_image_formation,
            RMA=init_rma)

    def _populate_basic_image_data(self,
                                   sicd: SICDType,
                                   grid_node: Optional[ElementTree.Element]) -> None:
        """
        Populate the basic ImageData and GeoData. This assumes not ScanSAR mode.
        This modifies the provided sicd in place.

        Parameters
        ----------
        sicd : SICDType
        grid_node : None|ElementTree.Element
            The central geolocationGrid point, if it exists.

        Returns
        -------
        None
        """

        # NB: the role of rows and columns is switched in TSX/SICD convention
        rows = int(self._find_main('./productInfo/imageDataInfo/imageRaster/numberOfColumns').text)
        cols = int(self._find_main('./productInfo/imageDataInfo/imageRaster/numberOfRows').text)

        if grid_node is not None:
            scp_row = int(grid_node.find('./col').text)
            scp_col = int(grid_node.find('./row').text)
            scp_llh = [
                float(grid_node.find('./lat').text),
                float(grid_node.find('./lon').text),
                float(grid_node.find('./height').text)]
        else:
            scp_row = int(self._find_main('./productInfo/sceneInfo/sceneCenterCoord/refColumn').text) - 1
            scp_col = int(self._find_main('./productInfo/sceneInfo/sceneCenterCoord/refRow').text) - 1
            scp_llh = [
                float(self._find_main('./productInfo/sceneInfo/sceneCenterCoord/lat').text),
                float(self._find_main('./productInfo/sceneInfo/sceneCenterCoord/lon').text),
                float(self._find_main('./productInfo/sceneInfo/sceneAverageHeight').text)]

        if self._find_main('./productInfo/acquisitionInfo/lookDirection').text[0].upper() == 'L':
            scp_col = cols - scp_col - 1

        sicd.ImageData = ImageDataType(
            NumRows=rows, NumCols=cols, FirstRow=0, FirstCol=0, FullImage=(rows, cols),
            PixelType='RE16I_IM16I', SCPPixel=(scp_row, scp_col))
        sicd.GeoData = GeoDataType(SCP=SCPType(LLH=scp_llh))

    @staticmethod
    def _populate_initial_radar_collection(
            sicd: SICDType,
            tx_pols: List[str],
            tx_rcv_pols: List[str]) -> None:
        """
        Populate the initial radar collection information. This modifies the
        provided sicd in place.

        Parameters
        ----------
        sicd : SICDType
        tx_pols : List[str]
        tx_rcv_pols : List[str]

        Returns
        -------
        None
        """

        tx_pol_count = len(set(tx_pols))
        if tx_pol_count == 1:
            the_tx_pol = tx_pols[0]
            tx_sequence = None
        else:
            the_tx_pol = 'SEQUENCE'
            tx_sequence = [TxStepType(TxPolarization=tx_pol) for tx_pol in tx_pols]
        sicd.RadarCollection = RadarCollectionType(
            TxPolarization=the_tx_pol,
            TxSequence=tx_sequence,
            RcvChannels=[ChanParametersType(TxRcvPolarization=tx_rcv_pol) for tx_rcv_pol in tx_rcv_pols])

    def _complete_sicd(self,
                       sicd: SICDType,
                       orig_pol: str,
                       layer_index: str,
                       pol_index: int,
                       ss_zd_s: float,
                       side_of_track: str,
                       center_freq: float,
                       arp_times: numpy.ndarray,
                       arp_pos: numpy.ndarray,
                       arp_vel: numpy.ndarray,
                       middle_grid: Optional[ElementTree.Element],
                       doppler_rate_reference_node: ElementTree.Element) -> SICDType:
        """
        Complete the remainder of the sicd information and populate as collection,
        if appropriate. **This assumes that this is not ScanSAR mode.**

        Parameters
        ----------
        sicd : SICDType
        orig_pol : str
            The TSX polarization string.
        layer_index : str
            The layer index entry.
        pol_index : int
            The polarization index (1 based) here.
        ss_zd_s : float
            The zero doppler spacing in the time domain.
        side_of_track : str
            One of ['R', 'S']
        center_freq : float
            The center frequency.
        arp_times : numpy.ndarray
            The array of reference times for the state information.
        arp_pos : numpy.ndarray
        arp_vel : numpy.ndarray
        middle_grid : None|ElementTree.Element
            The central geolocationGrid point, if it exists.
        doppler_rate_reference_node : ElementTree.Element

        Returns
        -------
        SICDType
        """

        def get_settings_node():
            # type: () -> Union[None, ElementTree.Element]
            for entry in self._findall_main('./instrument/settings'):
                if entry.find('./polLayer').text == orig_pol:
                    return entry
            return None

        def set_timeline():
            prf = float(settings_node.find('./settingRecord/PRF').text)
            ipp_poly = Poly1DType(Coefs=[0, prf])
            out_sicd.Timeline = TimelineType(
                CollectStart=collect_start,
                CollectDuration=collect_duration,
                IPP=[IPPSetType(TStart=0,
                                TEnd=collect_duration,
                                IPPPoly=ipp_poly,
                                IPPStart=0,
                                IPPEnd=int(round(ipp_poly(collect_duration))) - 1)])

        def set_position():
            times_s = numpy.array(
                [get_seconds(entry, collect_start, precision='us') for entry in arp_times], dtype='float64')
            P_x, P_y, P_z = fit_position_xvalidation(times_s, arp_pos, arp_vel, max_degree=8)
            out_sicd.Position = PositionType(ARPPoly=XYZPolyType(X=P_x, Y=P_y, Z=P_z))

        def complete_radar_collection():
            tx_pulse_length = float(
                self._find_main('./processing'
                                '/processingParameter'
                                '/rangeCompression'
                                '/chirps'
                                '/referenceChirp'
                                '/pulseLength').text)*32/3.29658384e8
            # NB: the matlab version indicates that this conversion comes via personal
            # communication with Juergen Janoth, Head of Application Development, Infoterra
            # The times of this communication is not indicated

            sample_rate = float(settings_node.find('./RSF').text)
            rcv_window_length = float(settings_node.find('./settingRecord/echowindowLength').text)/sample_rate

            out_sicd.RadarCollection.TxFrequency = (tx_freq_start, tx_freq_end)
            out_sicd.RadarCollection.Waveform = [
                WaveformParametersType(TxPulseLength=tx_pulse_length,
                                       TxRFBandwidth=band_width,
                                       TxFreqStart=tx_freq_start,
                                       TxFMRate=band_width/tx_pulse_length,
                                       ADCSampleRate=sample_rate,
                                       RcvWindowLength=rcv_window_length,
                                       RcvFMRate=0)]

        def complete_image_formation():
            out_sicd.ImageFormation.RcvChanProc.ChanIndices = [pol_index, ]
            out_sicd.ImageFormation.TEndProc = collect_duration
            out_sicd.ImageFormation.TxFrequencyProc = (tx_freq_start, tx_freq_end)
            out_sicd.ImageFormation.TxRcvPolarizationProc = self._get_sicd_tx_rcv_pol(orig_pol)

        def complete_rma():
            use_ss_zd_s = -ss_zd_s if side_of_track == 'L' else ss_zd_s
            time_ca_linear = use_ss_zd_s/out_sicd.Grid.Col.SS
            if self._georef_root is not None:
                if middle_grid is None:
                    raise ValueError('middle_grid should have been provided here')

                ref_time = parse_timestring(
                    self._find_georef('./geolocationGrid'
                                      '/gridReferenceTime'
                                      '/tReferenceTimeUTC').text, precision='us')
                az_offset = get_seconds(ref_time, collect_start, precision='us')
                time_ca_scp = float(middle_grid.find('./t').text)
                # get the sum of all provided azimuth shifts?
                # NB: this is obviously assuming that all entries are constant shifts...should we check?
                azimuths_shifts = [
                    float(entry.find('./coefficient').text) for entry in
                    self._findall_georef('./signalPropagationEffects/azimuthShift')]
                azimuth_shift = reduce(lambda x, y: x+y, azimuths_shifts)
                out_sicd.RMA.INCA.TimeCAPoly = Poly1DType(
                    Coefs=[time_ca_scp + az_offset - azimuth_shift, time_ca_linear])
                azimuth_time_scp = get_seconds(ref_time, collect_start, precision='us') + time_ca_scp

                range_time_scp = float(
                    self._find_georef('./geolocationGrid/gridReferenceTime/tauReferenceTime').text) + \
                    float(middle_grid.find('./tau').text)
                # get the sum of all provided range delays?
                # NB: this is obviously assuming that all entries are constant shifts...should we check?
                range_delays = [
                    float(entry.find('./coefficient').text) for entry in
                    self._findall_georef('./signalPropagationEffects/rangeDelay')]
                range_delay = reduce(lambda x, y: x+y, range_delays)
                out_sicd.RMA.INCA.R_CA_SCP = 0.5*(range_time_scp - range_delay)*speed_of_light
            else:
                azimuth_time_scp = get_seconds(
                    parse_timestring(
                        self._find_main('./productInfo/sceneInfo/sceneCenterCoord/azimuthTimeUTC').text,
                        precision='us'),
                    collect_start, precision='us')
                range_time_scp = float(self._find_main('./productInfo/sceneInfo/sceneCenterCoord/rangeTime').text)
                out_sicd.RMA.INCA.TimeCAPoly = Poly1DType(Coefs=[azimuth_time_scp, time_ca_linear])
                out_sicd.RMA.INCA.R_CA_SCP = 0.5*range_time_scp*speed_of_light

            # populate DopCentroidPoly and TimeCOAPoly
            if out_sicd.CollectionInfo.RadarMode.ModeID == 'ST':
                # proper spotlight mode
                coa_time = collect_duration / 2
                look = 1 if side_of_track == 'L' else -1
                alpha = 2.0/speed_of_light
                pos = out_sicd.Position.ARPPoly(coa_time)
                vel = out_sicd.Position.ARPPoly.derivative_eval(coa_time)
                speed = numpy.linalg.norm(vel)
                vel_hat = vel / speed
                los = out_sicd.GeoData.SCP.ECF.get_array() - pos

                out_sicd.Grid.TimeCOAPoly = Poly2DType(Coefs=[[coa_time, ], ])
                dop_poly = numpy.zeros((2, 2), dtype=numpy.float64)
                dop_poly[0, 1] = -look*center_freq*alpha*speed/out_sicd.RMA.INCA.R_CA_SCP
                dop_poly[1, 1] = look*center_freq*alpha*speed/(out_sicd.RMA.INCA.R_CA_SCP**2)
                dop_poly[:, 0] = dop_poly[:, 0] - look*(dop_poly[:, 1]*numpy.dot(los, vel_hat))
                out_sicd.Grid.Col.DeltaKCOAPoly = Poly2DType(Coefs=dop_poly*use_ss_zd_s/out_sicd.Grid.Col.SS)
            else:
                dop_centroid_poly, time_coa_poly = self._calculate_dop_polys(
                    layer_index, azimuth_time_scp, range_time_scp, collect_start, doppler_rate_reference_node)
                out_sicd.RMA.INCA.DopCentroidPoly = Poly2DType(Coefs=dop_centroid_poly)
                out_sicd.RMA.INCA.DopCentroidCOA = True
                out_sicd.Grid.TimeCOAPoly = Poly2DType(Coefs=time_coa_poly)
                out_sicd.Grid.Col.DeltaKCOAPoly = Poly2DType(Coefs=dop_centroid_poly*use_ss_zd_s/out_sicd.Grid.Col.SS)
            # calculate DRateSFPoly
            vm_vel_sq = numpy.sum(out_sicd.Position.ARPPoly.derivative_eval(azimuth_time_scp)**2)
            r_ca = numpy.array([out_sicd.RMA.INCA.R_CA_SCP, 1], dtype='float64')
            dop_rate_poly_coefs = [
                float(entry.text) for entry in doppler_rate_reference_node.findall(
                    './dopplerRatePolynomial/coefficient')]
            # Shift 1D polynomial to account for SCP
            dop_rate_ref_time = float(doppler_rate_reference_node.find('./dopplerRatePolynomial/referencePoint').text)
            dop_rate_poly_rg = Poly1DType(Coefs=dop_rate_poly_coefs).shift(dop_rate_ref_time - range_time_scp,
                                                                           alpha=2/speed_of_light,
                                                                           return_poly=False)

            # NB: assumes a sign of -1
            drate_poly = -polynomial.polymul(dop_rate_poly_rg, r_ca)*speed_of_light/(2*center_freq*vm_vel_sq)
            out_sicd.RMA.INCA.DRateSFPoly = Poly2DType(Coefs=numpy.reshape(drate_poly, (-1, 1)))

        def define_radiometric():
            beta_factor = float(self._find_main('./calibration'
                                                '/calibrationConstant[@layerIndex="{}"]'.format(layer_index) +
                                                '/calFactor').text)
            range_time_scp = float(self._find_main('./productInfo/sceneInfo/sceneCenterCoord/rangeTime').text)
            # now, calculate the radiometric noise polynomial
            # find the noise node
            noise_node = self._find_main('./noise[@layerIndex="{}"]'.format(layer_index))
            # extract the middle image noise node
            noise_data_nodes = noise_node.findall('./imageNoise')
            noise_data_node = noise_data_nodes[int(len(noise_data_nodes)/2)]
            range_min = float(noise_data_node.find('./noiseEstimate/validityRangeMin').text)
            range_max = float(noise_data_node.find('./noiseEstimate/validityRangeMax').text)
            ref_point = float(noise_data_node.find('./noiseEstimate/referencePoint').text)
            poly_coeffs = numpy.array(
                [float(coeff.text) for coeff in noise_data_node.findall(
                    './noiseEstimate/coefficient')], dtype='float64')
            # create a sample grid in range time and evaluate the noise
            range_time = numpy.linspace(range_min, range_max, 100) - ref_point
            # this should be an absolute squared magnitude value
            raw_noise_values = polynomial.polyval(range_time, poly_coeffs)
            # we convert to db
            noise_values = 10*numpy.log10(raw_noise_values)
            coords_range_m = 0.5*(range_time + ref_point - range_time_scp)*speed_of_light
            # fit the polynomial
            scale = 1e-3
            deg = poly_coeffs.size-1
            coeffs = polynomial.polyfit(coords_range_m*scale, noise_values, deg=deg, rcond=1e-30, full=False)
            coeffs *= numpy.power(scale, numpy.arange(deg+1))
            coeffs = numpy.reshape(coeffs, (-1, 1))
            out_sicd.Radiometric = RadiometricType(
                BetaZeroSFPoly=Poly2DType(Coefs=[[beta_factor, ], ]),
                NoiseLevel=NoiseLevelType_(
                    NoiseLevelType='ABSOLUTE', NoisePoly=Poly2DType(Coefs=coeffs)))

        def revise_scp():
            scp_ecf = out_sicd.project_image_to_ground(out_sicd.ImageData.SCPPixel.get_array())
            out_sicd.update_scp(scp_ecf, coord_system='ECF')

        out_sicd = sicd.copy()
        # get some common use parameters
        settings_node = get_settings_node()
        if settings_node is None:
            raise ValueError('Cannot find the settings node for polarization {}'.format(orig_pol))
        collect_start = parse_timestring(
            settings_node.find('./settingRecord'
                               '/dataSegment'
                               '/startTimeUTC').text, precision='us')
        collect_end = parse_timestring(
            settings_node.find('./settingRecord'
                               '/dataSegment'
                               '/stopTimeUTC').text, precision='us')
        collect_duration = get_seconds(collect_end, collect_start, precision='us')
        band_width = float(settings_node.find('./rxBandwidth').text)
        tx_freq_start = center_freq - 0.5 * band_width
        tx_freq_end = center_freq + 0.5 * band_width

        # populate the missing sicd elements
        set_timeline()
        set_position()
        complete_radar_collection()
        complete_image_formation()
        complete_rma()
        define_radiometric()
        out_sicd.derive()
        revise_scp()
        out_sicd.populate_rniirs(override=False)
        return out_sicd

    def get_sicd_collection(self) -> Tuple[List[str], List[SICDType]]:
        """
        Gets the sicd metadata collection.

        Returns
        -------
        files: List[str]
        sicds: List[SICDType]
        """

        def get_file_name(layer_index):
            file_node = self._find_main(
                './productComponents/imageData[@layerIndex="{}"]/file/location'.format(layer_index))
            path_stem = file_node.find('./path').text
            file_name = file_node.find('./filename').text
            full_file = os.path.join(self._parent_directory, path_stem, file_name)
            if not os.path.isfile(full_file):
                raise ValueError('Expected image file at\n\t{}\n\tbut this path does not exist'.format(full_file))
            return full_file

        the_files = []
        the_sicds = []

        # get some basic common use parameters
        center_freq = float(self._find_main('./instrument/radarParameters/centerFrequency').text)
        dop_bw = float(self._find_main('./processing/processingParameter/azimuthLookBandwidth').text)
        ss_zd_s = float(self._find_main('./productInfo/imageDataInfo/imageRaster/columnSpacing').text)
        side_of_track = self._find_main('./productInfo/acquisitionInfo/lookDirection').text[0].upper()

        # define the basic SICD shell
        basic_sicd = self._get_basic_sicd_shell(center_freq, dop_bw, ss_zd_s)
        if basic_sicd.CollectionInfo.RadarMode.ModeID == 'SC':
            raise ValueError('ScanSAR mode is currently unsupported')

        # fetch the state vector data
        times, positions, velocities = self._get_state_vector_data()
        # fetch the polarization list(s) (maybe ScanSAR modification required here)
        original_pols, tx_pols, tx_rcv_pols = self._get_full_pol_list()
        if basic_sicd.CollectionInfo.RadarMode.ModeID == 'SC':
            raise ValueError('ScanSAR mode is currently unsupported')
        else:
            middle_grid = self._find_middle_grid_node()
            self._populate_basic_image_data(basic_sicd, middle_grid)
            self._populate_initial_radar_collection(basic_sicd, tx_pols, tx_rcv_pols)
            # get the doppler rate reference node
            doppler_rate_nodes = self._findall_main('./processing/geometry/dopplerRate')
            doppler_rate_center_node = doppler_rate_nodes[int(len(doppler_rate_nodes) / 2)]

            for i, orig_pol in enumerate(original_pols):
                the_layer = '{}'.format(i+1)
                pol_index = i+1
                the_sicds.append(self._complete_sicd(
                    basic_sicd, orig_pol, the_layer, pol_index, ss_zd_s, side_of_track,
                    center_freq, times, positions, velocities, middle_grid, doppler_rate_center_node))
                the_files.append(get_file_name(the_layer))
        return the_files, the_sicds


class COSARDetails(object):
    __slots__ = (
        '_file_name', '_file_size', '_header_offsets', '_data_offsets',
        '_burst_index', '_burst_size', '_data_sizes', '_version')

    def __init__(self, file_name: str):
        """

        Parameters
        ----------
        file_name : str
        """

        self._header_offsets = []
        self._data_offsets = []
        self._burst_index = []
        self._burst_size = []
        self._data_sizes = []
        self._version = None

        if not os.path.isfile(file_name):
            raise SarpyIOError('path {} is not not a file'.format(file_name))
        self._file_name = file_name
        self._file_size = os.path.getsize(file_name)
        self._parse_details()

    @property
    def burst_count(self) -> int:
        """
        int: The discovered burst count
        """

        return len(self._data_offsets)

    @property
    def version(self) -> Optional[int]:
        """
        int: The COSAR version
        """

        return self._version

    @property
    def pixel_type(self) -> Optional[str]:
        """
        str: The pixel type
        """

        if self._version is None:
            return None
        elif self._version == 1:
            return 'RE16I_IM16I'
        elif self._version == 2:
            return 'RE32F_IM32F'
        else:
            raise ValueError('Got unexpected version {}'.format(self._version))

    def _process_burst_header(
            self,
            fi: BinaryIO,
            the_offset: int):
        """

        Parameters
        ----------
        fi : BinaryIO
        the_offset : int

        Returns
        -------
        None
        """

        if the_offset >= self._file_size - 48:
            raise ValueError(
                'The seek location + basic header size is greater than the file size.')
        # seek to our desired location
        fi.seek(the_offset, os.SEEK_SET)
        # read the desired bytes
        header_bytes = fi.read(48)
        # interpret the data
        burst_in_bytes = struct.unpack('>I', header_bytes[:4])[0]
        rsri = struct.unpack('>I', header_bytes[4:8])[0]
        range_samples = struct.unpack('>I', header_bytes[8:12])[0]
        azimuth_samples = struct.unpack('>I', header_bytes[12:16])[0]
        burst_index = struct.unpack('>I', header_bytes[16:20])[0]
        # these two are only useful in the first record
        rtnb = struct.unpack('>I', header_bytes[20:24])[0]
        tnl = struct.unpack('>I', header_bytes[24:28])[0]
        # basic check bytes
        csar = struct.unpack('>4s', header_bytes[28:32])[0]
        version = struct.unpack('>I', header_bytes[32:36])[0]
        oversample = struct.unpack('>I', header_bytes[36:40])[0]
        scaling_rate = struct.unpack('>d', header_bytes[40:])[0]
        if csar.upper() != b'CSAR':
            raise ValueError('unexpected csar value {}'.format(csar))
        logger.debug(
            'Parsed COSAR burst:'
            '\n\tburst_in_bytes = {}'
            '\n\trsri = {}'
            '\n\trange samples = {}'
            '\n\tazimuth samples = {}'
            '\n\trtnb = {}'
            '\n\ttnl = {}'
            '\n\tcsar = {}'
            '\n\tversion = {}'
            '\n\toversample = {}'
            '\n\tscaling rate = {}'.format(
                burst_in_bytes, rsri, range_samples, azimuth_samples, rtnb, tnl,
                csar, version, oversample, scaling_rate))

        # now, populate our appropriate details
        data_offset = the_offset + (int(range_samples)+2)*4*4
        burst_size = 4*(int(range_samples)+2)*(int(azimuth_samples) + 4)
        self._header_offsets.append(the_offset)
        self._data_offsets.append(data_offset)
        self._burst_index.append(int(burst_index))
        self._burst_size.append(burst_size)
        self._data_sizes.append((range_samples, azimuth_samples))
        self._version = version
        if the_offset + burst_size > self._file_size:
            raise ValueError(
                'The file size for {} is given as {} bytes, but '
                'the burst at index {} has size {} and offset {}'.format(
                    self._file_name, self._file_size, self._burst_index[-1],
                    self._burst_size[-1], the_offset))
        if self._version not in [1, 2]:
            raise ValueError(
                'Got unexpected version value {}'.format(self._version))

    def _parse_details(self) -> None:
        with open(self._file_name, 'rb') as fi:
            # process the first burst header
            self._process_burst_header(fi, 0)
            cont = True
            while cont:
                next_burst_location = self._header_offsets[-1] + self._burst_size[-1]
                if next_burst_location < self._file_size:
                    self._process_burst_header(fi, next_burst_location)
                else:
                    cont = False

    def construct_data_segment(
            self,
            index: int,
            reverse_axes: Optional[Tuple[int, ...]],
            transpose_axes: Optional[Tuple[int, ...]],
            expected_size: Tuple[int, ...]) -> DataSegment:
        """
        Construct a data segment for the given burst index.

        Parameters
        ----------
        index : int
        reverse_axes : None|int|Sequence[int, ...]
        transpose_axes : None|Tuple[int, ...]
        expected_size : Tuple[int, ...]

        Returns
        -------
        DataSegment
        """

        index = int(index)
        if not (0 <= index < self.burst_count):
            raise KeyError('Provided index {} must be in the range [0, {})'.format(index, self.burst_count))
        # get data_size
        offset = self._data_offsets[index]
        range_samples, azimuth_samples = self._data_sizes[index]
        exp_cols, exp_rows = expected_size
        if not (exp_rows == range_samples and exp_cols == azimuth_samples):
            raise ValueError(
                'Expected raw burst size is {}, while actual raw burst size '
                'is {}'.format(expected_size, (range_samples, azimuth_samples)))

        if self._version == 1:
            raw_dtype = numpy.dtype('>i2')
        elif self._version == 2:
            raw_dtype = numpy.dtype('>f2')
        else:
            raise ValueError('Got unhandled version {}'.format(self._version))

        format_function = ComplexFormatFunction(raw_dtype, 'IQ')
        parent_segment = NumpyMemmapSegment(
            self._file_name, offset, raw_dtype, (azimuth_samples, range_samples + 2, 2),
            'complex64', (range_samples + 2, azimuth_samples),
            reverse_axes=reverse_axes, transpose_axes=transpose_axes,
            format_function=format_function, mode='r', close_file=True)
        subset_def = (slice(2, exp_rows+2), slice(0, exp_cols))
        relevant_segment = SubsetSegment(parent_segment, subset_def, 'formatted', close_parent=True)
        return relevant_segment


#########
# the reader implementation

class TSXReader(SICDTypeReader):
    """
    The TerraSAR-X SLC reader implementation.

    **Changed in version 1.3.0** for reading changes.
    """

    __slots__ = ('_tsx_details', )

    def __init__(self, tsx_details):
        """

        Parameters
        ----------
        tsx_details : str|TSXDetails
        """

        if isinstance(tsx_details, str):
            tsx_details = TSXDetails(tsx_details)
        if not isinstance(tsx_details, TSXDetails):
            raise TypeError(
                'tsx_details is expected to be the path to the TerraSAR-X package '
                'directory or main xml file, of TSXDetails instance. Got type {}'.format(type(tsx_details)))
        self._tsx_details = tsx_details

        data_segments = []
        image_format = tsx_details.image_format
        the_files, the_sicds = tsx_details.get_sicd_collection()
        for the_file, the_sicd in zip(the_files, the_sicds):
            rows = the_sicd.ImageData.NumRows
            cols = the_sicd.ImageData.NumCols
            reverse_axes = (0, ) if the_sicd.SCPCOA.SideOfTrack == 'L' else None
            transpose_axes = (1, 0, 2)
            if image_format != 'COSAR':
                raise ValueError(
                    'Expected complex data for TerraSAR-X to be in COSAR format. '
                    'Got unhandled format {}'.format(image_format))
            cosar_details = COSARDetails(the_file)
            if cosar_details.burst_count != 1:
                raise ValueError(
                    'Expected one burst in the COSAR file {},\n\t'
                    'but got {} bursts'.format(the_file, cosar_details.burst_count))
            the_sicd.ImageData.PixelType = cosar_details.pixel_type
            data_seg = cosar_details.construct_data_segment(0, reverse_axes, transpose_axes, (cols, rows))
            data_segments.append(data_seg)
        SICDTypeReader.__init__(self, data_segments, the_sicds, close_segments=True)
        self._check_sizes()

    @property
    def file_name(self) -> str:
        return self._tsx_details.file_name


########
# base expected functionality for a module with an implemented Reader

def is_a(file_name: str) -> Optional[TSXReader]:
    """
    Tests whether a given file_name corresponds to a TerraSAR-X file SSC package.
    Returns a reader instance, if so.

    Parameters
    ----------
    file_name : str
        the file_name to check

    Returns
    -------
    TSXReader|None
        `TSXReader` instance if TerraSAR-X file file, `None` otherwise
    """

    if is_file_like(file_name):
        return None

    try:
        tsx_details = TSXDetails(file_name)
        logger.info('Path {} is determined to be a TerraSAR-X file package.'.format(tsx_details.file_name))
        return TSXReader(tsx_details)
    except SarpyIOError:
        return None



"""
Methods for creating a variety of SIDD products.

Examples
--------
Create a variety of sidd products.

.. code-block:: python

    import os

    from sarpy.io.complex.converter import open_complex
    from sarpy.processing.ortho_rectify import BivariateSplineMethod, NearestNeighborMethod
    from sarpy.processing.sidd.sidd_product_creation import create_detected_image_sidd, create_csi_sidd, create_dynamic_image_sidd

    # open a sicd type file
    reader = open_complex('<sicd type object file name>')
    # create an orthorectification helper for specified sicd index
    ortho_helper = NearestNeighborMethod(reader, index=0)

    # create a sidd version 2 detected image for the whole file
    create_detected_image_sidd(ortho_helper, '<output directory>', block_size=10, version=2)
    # create a sidd version 2 color sub-aperture image for the whole file
    create_csi_sidd(ortho_helper, '<output directory>', dimension=0, version=2)
    # create a sidd version 2 dynamic image/sub-aperture stack for the whole file
    create_dynamic_image_sidd(ortho_helper, '<output directory>', dimension=0, version=2)
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import os
from sarpy.processing.ortho_rectify.base import FullResolutionFetcher, OrthorectificationIterator
from sarpy.processing.ortho_rectify.ortho_methods import OrthorectificationHelper
from sarpy.processing.sidd.sidd_structure_creation import create_sidd_structure
from sarpy.processing.sicd.csi import CSICalculator
from sarpy.processing.sicd.subaperture import SubapertureCalculator, SubapertureOrthoIterator
from sarpy.io.product.sidd import SIDDWriter
from sarpy.io.general.base import SarpyIOError
from sarpy.visualization.remap import MonochromaticRemap, NRL

# TODO: move this to processing for 1.3.0

DEFAULT_IMG_REMAP = NRL
DEFAULT_CSI_REMAP = NRL
DEFAULT_DI_REMAP = NRL

_output_text = 'output_directory `{}`\n\t' \
               'does not exist or is not a directory'
_orthohelper_text = 'ortho_helper is required to be an instance of OrthorectificationHelper,\n\t' \
                    'got type `{}`'


def _validate_filename(output_directory, output_file, sidd_structure):
    """
    Validate the output filename.

    Parameters
    ----------
    output_directory : str
    output_file : None|str
    sidd_structure

    Returns
    -------
    str
    """

    if output_file is None:
        # noinspection PyProtectedMember
        fstem = os.path.split(sidd_structure.NITF['SUGGESTED_NAME']+'.nitf')[1]
    else:
        fstem = os.path.split(output_file)[1]

    full_filename = os.path.join(os.path.expanduser(output_directory), fstem)
    if os.path.exists(full_filename):
        raise SarpyIOError('File {} already exists.'.format(full_filename))
    return full_filename


def _validate_remap_function(remap_function):
    """
    Verify that the given monochromatic remap function is viable for SIDD
    production.

    Parameters
    ----------
    remap_function : MonochromaticRemap
    """

    if not isinstance(remap_function, MonochromaticRemap):
        raise TypeError('remap_function must be an instance of MonochromaticRemap')
    if remap_function.bit_depth not in [8, 16]:
        raise TypeError('remap_function usage for SIDD requires 8 or 16 bit output')


def create_detected_image_sidd(
        ortho_helper, output_directory, output_file=None, block_size=10, dimension=0,
        bounds=None, version=3, include_sicd=True, remap_function=None):
    """
    Create a SIDD version of a basic detected image from a SICD type reader.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
        The ortho-rectification helper object.
    output_directory : str
        The output directory for the given file.
    output_file : None|str
        The file name, this will default to a sensible value.
    block_size : int
        The approximate processing block size to fetch, given in MB. The
        minimum value for use here will be 1.
    dimension : int
        Which dimension to split over in block processing? Must be either 0 or 1.
    bounds : None|numpy.ndarray|list|tuple
        The sicd pixel bounds of the form `(min row, max row, min col, max col)`.
        This will default to the full image.
    version : int
        The SIDD version to use, must be one of 1, 2, or 3.
    include_sicd : bool
        Include the SICD structure in the SIDD file?
    remap_function : None|MonochromaticRemap
        The applied remap function. If one is not provided, then a default is
        used. Required global parameters will be calculated if they are missing,
        so the internal state of this remap function may be modified.

    Returns
    -------
    None

    Examples
    --------
    .. code-block:: python

        import os

        from sarpy.io.complex.converter import open_complex
        from sarpy.processing.ortho_rectify import BivariateSplineMethod, NearestNeighborMethod
        from sarpy.processing.sidd.sidd_product_creation import create_detected_image_sidd

        reader = open_complex('<sicd type object file name>')
        ortho_helper = NearestNeighborMethod(reader, index=0)

        # create a sidd version 2 file for the whole file
        create_detected_image_sidd(ortho_helper, '<output directory>', block_size=10, version=2)
    """

    if not os.path.isdir(output_directory):
        raise SarpyIOError(_output_text.format(output_directory))

    if not isinstance(ortho_helper, OrthorectificationHelper):
        raise TypeError(_orthohelper_text.format(type(ortho_helper)))

    if remap_function is None:
        remap_function = DEFAULT_IMG_REMAP(override_name='IMG_DEFAULT')
    _validate_remap_function(remap_function)

    # construct the ortho-rectification iterator - for a basic data fetcher
    calculator = FullResolutionFetcher(
        ortho_helper.reader, dimension=dimension, index=ortho_helper.index, block_size=block_size)
    ortho_iterator = OrthorectificationIterator(
        ortho_helper, calculator=calculator, bounds=bounds,
        remap_function=remap_function, recalc_remap_globals=False)

    # create the sidd structure
    ortho_bounds = ortho_iterator.ortho_bounds
    sidd_structure = create_sidd_structure(
        ortho_helper, ortho_bounds,
        product_class='Detected Image', pixel_type='MONO{}I'.format(remap_function.bit_depth), version=version)
    # set suggested name
    sidd_structure.NITF['SUGGESTED_NAME'] = ortho_helper.sicd.get_suggested_name(ortho_helper.index)+'_IMG'

    # create the sidd writer
    full_filename = _validate_filename(output_directory, output_file, sidd_structure)
    with SIDDWriter(full_filename, sidd_structure, ortho_helper.sicd if include_sicd else None) as writer:
        # iterate and write
        for data, start_indices in ortho_iterator:
            writer(data, start_indices=start_indices, index=0)


def create_csi_sidd(
        ortho_helper, output_directory, output_file=None, dimension=0,
        block_size=30, bounds=None, version=3, include_sicd=True, remap_function=None):
    """
    Create a SIDD version of a Color Sub-Aperture Image from a SICD type reader.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
        The ortho-rectification helper object.
    output_directory : str
        The output directory for the given file.
    output_file : None|str
        The file name, this will default to a sensible value.
    dimension : int
        The dimension over which to split the sub-aperture.
    block_size : int
        The approximate processing block size to fetch, given in MB. The
        minimum value for use here will be 1.
    bounds : None|numpy.ndarray|list|tuple
        The sicd pixel bounds of the form `(min row, max row, min col, max col)`.
        This will default to the full image.
    version : int
        The SIDD version to use, must be one of 1, 2, or 3.
    include_sicd : bool
        Include the SICD structure in the SIDD file?
    remap_function : None|MonochromaticRemap
        The applied remap function. For csi processing, this must explicitly be
        an 8-bit remap. If one is not provided, then a default is used. Required
        global parameters will be calculated if they are missing, so the internal
        state of this remap function may be modified.

    Returns
    -------
    None

    Examples
    --------
    .. code-block:: python

        import os
        from sarpy.io.complex.converter import open_complex
        from sarpy.processing.sidd.sidd_product_creation import create_csi_sidd
        from sarpy.processing.sicd.csi import CSICalculator
        from sarpy.processing.ortho_rectify import NearestNeighborMethod

        reader = open_complex('<sicd type object file name>')
        ortho_helper = NearestNeighborMethod(reader, index=0)
        create_csi_sidd(ortho_helper, '<output directory>', dimension=0, version=2)

    """

    if not os.path.isdir(output_directory):
        raise SarpyIOError(_output_text.format(output_directory))

    if not isinstance(ortho_helper, OrthorectificationHelper):
        raise TypeError(_orthohelper_text.format(type(ortho_helper)))

    # construct the CSI calculator class
    csi_calculator = CSICalculator(
        ortho_helper.reader, dimension=dimension, index=ortho_helper.index, block_size=block_size)

    if remap_function is None:
        remap_function = DEFAULT_CSI_REMAP(override_name='CSI_DEFAULT', bit_depth=8)
    _validate_remap_function(remap_function)
    if remap_function.bit_depth != 8:
        raise ValueError('The CSI SIDD specifically requires an 8-bit remap function.')

    # construct the ortho-rectification iterator
    ortho_iterator = OrthorectificationIterator(
        ortho_helper, calculator=csi_calculator, bounds=bounds,
        remap_function=remap_function, recalc_remap_globals=False)

    # create the sidd structure
    ortho_bounds = ortho_iterator.ortho_bounds
    sidd_structure = create_sidd_structure(
        ortho_helper, ortho_bounds,
        product_class='Color Subaperture Image', pixel_type='RGB24I', version=version)
    # set suggested name
    sidd_structure.NITF['SUGGESTED_NAME'] = csi_calculator.sicd.get_suggested_name(csi_calculator.index)+'_CSI'

    # create the sidd writer
    full_filename = _validate_filename(output_directory, output_file, sidd_structure)
    with SIDDWriter(full_filename, sidd_structure, csi_calculator.sicd if include_sicd else None) as writer:
        # iterate and write
        for data, start_indices in ortho_iterator:
            writer(data, start_indices=start_indices, index=0)


def create_dynamic_image_sidd(
        ortho_helper, output_directory, output_file=None, dimension=0, block_size=10,
        bounds=None, frame_count=9, aperture_fraction=0.2, method='FULL', version=3,
        include_sicd=True, remap_function=None):
    """
    Create a SIDD version of a Dynamic Image (Sub-Aperture Stack) from a SICD type reader.

    Parameters
    ----------
    ortho_helper : OrthorectificationHelper
        The ortho-rectification helper object.
    output_directory : str
        The output directory for the given file.
    output_file : None|str
        The file name, this will default to a sensible value.
    dimension : int
        The dimension over which to split the sub-aperture.
    block_size : int
        The approximate processing block size to fetch, given in MB. The
        minimum value for use here will be 1.
    bounds : None|numpy.ndarray|list|tuple
        The sicd pixel bounds of the form `(min row, max row, min col, max col)`.
        This will default to the full image.
    frame_count : int
        The number of frames to calculate.
    aperture_fraction : float
        The relative size of each aperture window.
    method : str
        The subaperture processing method, which must be one of
        `('NORMAL', 'FULL', 'MINIMAL')`.
    version : int
        The SIDD version to use, must be one of 1, 2, or 3.
    include_sicd : bool
        Include the SICD structure in the SIDD file?
    remap_function : None|MonochromaticRemap
        The applied remap function. If one is not provided, then a default is
        used. Required global parameters will be calculated if they are missing,
        so the internal state of this remap function may be modified.

    Returns
    -------
    None

    Examples
    --------
    Create a basic dynamic image.

    .. code-block:: python

        import os
        from sarpy.io.complex.converter import open_complex
        from sarpy.processing.sidd.sidd_product_creation import create_dynamic_image_sidd
        from sarpy.processing.sicd.csi import CSICalculator
        from sarpy.processing.ortho_rectify import NearestNeighborMethod

        reader = open_complex('<sicd type object file name>')
        ortho_helper = NearestNeighborMethod(reader, index=0)
        create_dynamic_image_sidd(ortho_helper, '<output directory>', dimension=0, version=2)
    """

    if not os.path.isdir(output_directory):
        raise SarpyIOError(_output_text.format(output_directory))

    if not isinstance(ortho_helper, OrthorectificationHelper):
        raise TypeError(_orthohelper_text.format(type(ortho_helper)))

    # construct the subaperture calculator class
    subap_calculator = SubapertureCalculator(
        ortho_helper.reader, dimension=dimension, index=ortho_helper.index, block_size=block_size,
        frame_count=frame_count, aperture_fraction=aperture_fraction, method=method)

    if remap_function is None:
        remap_function = DEFAULT_DI_REMAP(override_name='DI_DEFAULT')
    _validate_remap_function(remap_function)

    # construct the ortho-rectification iterator
    ortho_iterator = SubapertureOrthoIterator(
        ortho_helper, calculator=subap_calculator, bounds=bounds,
        remap_function=remap_function, recalc_remap_globals=False, depth_first=True)

    # create the sidd structure
    ortho_bounds = ortho_iterator.ortho_bounds
    sidd_structure = create_sidd_structure(
        ortho_helper, ortho_bounds,
        product_class='Dynamic Image', pixel_type='MONO{}I'.format(remap_function.bit_depth), version=version)
    # set suggested name
    sidd_structure.NITF['SUGGESTED_NAME'] = subap_calculator.sicd.get_suggested_name(subap_calculator.index)+'__DI'
    the_sidds = []
    for i in range(subap_calculator.frame_count):
        this_sidd = sidd_structure.copy()
        this_sidd.ProductCreation.ProductType = 'Frame {}'.format(i+1)
        the_sidds.append(this_sidd)

    # create the sidd writer
    if output_file is None:
        # noinspection PyProtectedMember
        full_filename = os.path.join(output_directory, sidd_structure.NITF['SUGGESTED_NAME']+'.nitf')
    else:
        full_filename = os.path.join(output_directory, output_file)
    if os.path.exists(os.path.expanduser(full_filename)):
        raise SarpyIOError('File {} already exists.'.format(full_filename))
    with SIDDWriter(full_filename, the_sidds, subap_calculator.sicd if include_sicd else None) as writer:
        # iterate and write
        for data, start_indices, the_frame in ortho_iterator:
            writer(data, start_indices=start_indices, index=the_frame)



"""
The MeasurementType definition.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from typing import Union, List

from sarpy.io.xml.base import Serializable, SerializableArray
from sarpy.io.xml.descriptors import SerializableDescriptor, UnitVectorDescriptor, \
    FloatDescriptor, StringEnumDescriptor, SerializableArrayDescriptor

from .base import DEFAULT_STRICT, FLOAT_FORMAT
from .blocks import ReferencePointType, RowColDoubleType, Poly2DType, XYZType, RowColIntType, \
    XYZPolyType, RowColArrayElement


class BaseProjectionType(Serializable):
    """
    The base geometric projection system.
    """

    _fields = ('ReferencePoint', )
    _required = ('ReferencePoint', )
    # Descriptor
    ReferencePoint = SerializableDescriptor(
        'ReferencePoint', ReferencePointType, _required, strict=DEFAULT_STRICT,
        docstring='Reference point for the geometrical system.')  # type: ReferencePointType

    def __init__(self, ReferencePoint=None, **kwargs):
        """

        Parameters
        ----------
        ReferencePoint : ReferencePointType
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.ReferencePoint = ReferencePoint
        super(BaseProjectionType, self).__init__(**kwargs)


class MeasurableProjectionType(BaseProjectionType):
    """
    A physical base projection.
    """

    _fields = ('ReferencePoint', 'SampleSpacing', 'TimeCOAPoly')
    _required = ('ReferencePoint', 'SampleSpacing', 'TimeCOAPoly')
    # Descriptor
    SampleSpacing = SerializableDescriptor(
        'SampleSpacing', RowColDoubleType, _required, strict=DEFAULT_STRICT,
        docstring='Sample spacing in row and column.')  # type: RowColDoubleType
    TimeCOAPoly = SerializableDescriptor(
        'TimeCOAPoly', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Time (units = seconds) at which center of aperture for a given pixel '
                  'coordinate in the product occurs.')  # type: Poly2DType

    def __init__(self, ReferencePoint=None, SampleSpacing=None, TimeCOAPoly=None, **kwargs):
        """

        Parameters
        ----------
        ReferencePoint : ReferencePointType
        SampleSpacing : RowColDoubleType|numpy.ndarray|list|tuple
        TimeCOAPoly : Poly2DType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        super(MeasurableProjectionType, self).__init__(ReferencePoint=ReferencePoint, **kwargs)
        self.SampleSpacing = SampleSpacing
        self.TimeCOAPoly = TimeCOAPoly


class ProductPlaneType(Serializable):
    """
    Plane definition for the product.
    """

    _fields = ('RowUnitVector', 'ColUnitVector')
    _required = _fields
    # Descriptor
    RowUnitVector = UnitVectorDescriptor(
        'RowUnitVector', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Unit vector of the plane defined to be aligned in the increasing row direction '
                  'of the product. (Defined as Rpgd in Design and Exploitation document)')  # type: XYZType
    ColUnitVector = UnitVectorDescriptor(
        'ColUnitVector', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Unit vector of the plane defined to be aligned in the increasing column direction '
                  'of the product. (Defined as Cpgd in Design and Exploitation document)')  # type: XYZType

    def __init__(self, RowUnitVector=None, ColUnitVector=None, **kwargs):
        """

        Parameters
        ----------
        RowUnitVector : XYZType|numpy.ndarray|list|tuple
        ColUnitVector : XYZType|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.RowUnitVector = RowUnitVector
        self.ColUnitVector = ColUnitVector
        super(ProductPlaneType, self).__init__(**kwargs)


class PlaneProjectionType(MeasurableProjectionType):
    """
    Planar representation of the pixel grid.
    """

    _fields = ('ReferencePoint', 'SampleSpacing', 'TimeCOAPoly', 'ProductPlane')
    _required = ('ReferencePoint', 'SampleSpacing', 'TimeCOAPoly', 'ProductPlane')
    # Descriptor
    ProductPlane = SerializableDescriptor(
        'ProductPlane', ProductPlaneType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: ProductPlaneType

    def __init__(self, ReferencePoint=None, SampleSpacing=None, TimeCOAPoly=None, ProductPlane=None, **kwargs):
        """

        Parameters
        ----------
        ReferencePoint : ReferencePointType
        SampleSpacing : RowColDoubleType|numpy.ndarray|list|tuple
        TimeCOAPoly : Poly2DType|numpy.ndarray|list|tuple
        ProductPlane : ProductPlaneType
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        super(PlaneProjectionType, self).__init__(
            ReferencePoint=ReferencePoint, SampleSpacing=SampleSpacing, TimeCOAPoly=TimeCOAPoly, **kwargs)
        self.ProductPlane = ProductPlane


class GeographicProjectionType(MeasurableProjectionType):
    """
    Geographic mapping of the pixel grid.
    """

    pass


class CylindricalProjectionType(MeasurableProjectionType):
    """
    Cylindrical mapping of the pixel grid.
    """

    _fields = ('ReferencePoint', 'SampleSpacing', 'TimeCOAPoly', 'StripmapDirection', 'CurvatureRadius')
    _required = ('ReferencePoint', 'SampleSpacing', 'TimeCOAPoly', 'StripmapDirection')
    _numeric_format = {'CurvatureRadius': FLOAT_FORMAT}
    # Descriptor
    StripmapDirection = SerializableDescriptor(
        'StripmapDirection', XYZType, _required, strict=DEFAULT_STRICT,
        docstring='Along stripmap direction.')  # type: XYZType
    CurvatureRadius = FloatDescriptor(
        'CurvatureRadius', _required, strict=DEFAULT_STRICT,
        docstring='Radius of Curvature defined at scene center.  If not present, the radius of '
                  'curvature will be derived based upon the equations provided in the '
                  'Design and Exploitation Document')  # type: Union[None, float]

    def __init__(self, ReferencePoint=None, SampleSpacing=None, TimeCOAPoly=None,
                 StripmapDirection=None, CurvatureRadius=None, **kwargs):
        """

        Parameters
        ----------
        ReferencePoint : ReferencePointType
        SampleSpacing : RowColDoubleType|numpy.ndarray|list|tuple
        TimeCOAPoly : Poly2DType|numpy.ndarray|list|tuple
        StripmapDirection : XYZType|numpy.ndarray|list|tuple
        CurvatureRadius : None|float
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        super(CylindricalProjectionType, self).__init__(
            ReferencePoint=ReferencePoint, SampleSpacing=SampleSpacing, TimeCOAPoly=TimeCOAPoly, **kwargs)
        self.StripmapDirection = StripmapDirection
        self.CurvatureRadius = CurvatureRadius


class PolynomialProjectionType(BaseProjectionType):
    """
    Polynomial pixel to ground.  This should only used for sensor systems where the radar
    geometry parameters are not recorded.
    """

    _fields = ('ReferencePoint', 'RowColToLat', 'RowColToLon', 'RowColToAlt', 'LatLonToRow', 'LatLonToCol')
    _required = ('ReferencePoint', 'RowColToLat', 'RowColToLon', 'LatLonToRow', 'LatLonToCol')
    # Descriptor
    RowColToLat = SerializableDescriptor(
        'RowColToLat', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial that converts Row/Col to Latitude (degrees).')  # type: Poly2DType
    RowColToLon = SerializableDescriptor(
        'RowColToLon', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial that converts Row/Col to Longitude (degrees).')  # type: Poly2DType
    RowColToAlt = SerializableDescriptor(
        'RowColToAlt', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial that converts Row/Col to Altitude (meters above '
                  'WGS-84 ellipsoid).')  # type: Union[None, Poly2DType]
    LatLonToRow = SerializableDescriptor(
        'LatLonToRow', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial that converts Latitude (degrees) and Longitude (degrees) to '
                  'pixel row location.')  # type: Poly2DType
    LatLonToCol = SerializableDescriptor(
        'LatLonToCol', Poly2DType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial that converts Latitude (degrees) and Longitude (degrees) to '
                  'pixel col location.')  # type: Poly2DType

    def __init__(self, ReferencePoint=None, RowColToLat=None, RowColToLon=None, RowColToAlt=None,
                 LatLonToRow=None, LatLonToCol=None, **kwargs):
        """

        Parameters
        ----------
        ReferencePoint : ReferencePointType
        RowColToLat : Poly2DType|numpy.ndarray|list|tuple
        RowColToLon : Poly2DType|numpy.ndarray|list|tuple
        RowColToAlt : None|Poly2DType|numpy.ndarray|list|tuple
        LatLonToRow : Poly2DType|numpy.ndarray|list|tuple
        LatLonToCol : Poly2DType|numpy.ndarray|list|tuple
        kwargs
        """
        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        super(PolynomialProjectionType, self).__init__(ReferencePoint=ReferencePoint, **kwargs)
        self.RowColToLat = RowColToLat
        self.RowColToLon = RowColToLon
        self.RowColToAlt = RowColToAlt
        self.LatLonToRow = LatLonToRow
        self.LatLonToCol = LatLonToCol


class MeasurementType(Serializable):
    """
    Geometric SAR information required for measurement/geolocation.
    """

    _fields = (
        'PolynomialProjection', 'GeographicProjection', 'PlaneProjection', 'CylindricalProjection',
        'PixelFootprint', 'ARPFlag', 'ARPPoly', 'ValidData')
    _required = ('PixelFootprint', 'ARPPoly', 'ValidData')
    _collections_tags = {'ValidData': {'array': True, 'child_tag': 'Vertex'}}
    _numeric_format = {'ValidData': FLOAT_FORMAT}
    _choice = ({'required': False, 'collection': ('PolynomialProjection', 'GeographicProjection',
                                                  'PlaneProjection', 'CylindricalProjection')}, )
    # Descriptor
    PolynomialProjection = SerializableDescriptor(
        'PolynomialProjection', PolynomialProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Polynomial pixel to ground. Should only used for sensor systems where the radar '
                  'geometry parameters are not recorded.')  # type: Union[None, PolynomialProjectionType]
    GeographicProjection = SerializableDescriptor(
        'GeographicProjection', GeographicProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Geographic mapping of the pixel grid referred to as GGD in the '
                  'Design and Exploitation document.')  # type: Union[None, GeographicProjectionType]
    PlaneProjection = SerializableDescriptor(
        'PlaneProjection', PlaneProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Planar representation of the pixel grid referred to as PGD in the '
                  'Design and Exploitation document.')  # type: Union[None, PlaneProjectionType]
    CylindricalProjection = SerializableDescriptor(
        'CylindricalProjection', CylindricalProjectionType, _required, strict=DEFAULT_STRICT,
        docstring='Cylindrical mapping of the pixel grid referred to as CGD in the '
                  'Design and Exploitation document.')  # type: Union[None, CylindricalProjectionType]
    PixelFootprint = SerializableDescriptor(
        'PixelFootprint', RowColIntType, _required, strict=DEFAULT_STRICT,
        docstring='Size of the image in pixels.')  # type: RowColIntType
    ARPFlag = StringEnumDescriptor(
        'ARPFlag', ('REALTIME', 'PREDICTED', 'POST PROCESSED'), _required, strict=DEFAULT_STRICT,
        docstring='Flag indicating whether ARP polynomial is based on the best available (`collect time` or '
                  '`predicted`) ephemeris.')  # type: Union[None, str]
    ARPPoly = SerializableDescriptor(
        'ARPPoly', XYZPolyType, _required, strict=DEFAULT_STRICT,
        docstring='')  # type: XYZPolyType
    ValidData = SerializableArrayDescriptor(
        'ValidData', RowColArrayElement, _collections_tags, _required, strict=DEFAULT_STRICT, minimum_length=3,
        docstring='Indicates the full image includes both valid data and some zero filled pixels. '
                  'Simple polygon encloses the valid data (may include some zero filled pixels for simplification). '
                  'Vertices in clockwise order.')  # type: Union[SerializableArray, List[RowColArrayElement]]

    def __init__(self, PolynomialProjection=None, GeographicProjection=None, PlaneProjection=None,
                 CylindricalProjection=None, PixelFootprint=None, ARPFlag=None, ARPPoly=None, ValidData=None, **kwargs):
        """

        Parameters
        ----------
        PolynomialProjection : PolynomialProjectionType
        GeographicProjection : GeographicProjectionType
        PlaneProjection : PlaneProjectionType
        CylindricalProjection : CylindricalProjectionType
        PixelFootprint : RowColIntType|numpy.ndarray|list|tuple
        ARPFlag : str
        ARPPoly : XYZPolyType|numpy.ndarray|list|tuple
        ValidData : SerializableArray|List[RowColArrayElement]|numpy.ndarray|list|tuple
        kwargs
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.PolynomialProjection = PolynomialProjection
        self.GeographicProjection = GeographicProjection
        self.PlaneProjection = PlaneProjection
        self.CylindricalProjection = CylindricalProjection
        self.PixelFootprint = PixelFootprint
        self.ARPFlag = ARPFlag
        self.ARPPoly = ARPPoly
        self.ValidData = ValidData
        super(MeasurementType, self).__init__(**kwargs)

    @property
    def ProjectionType(self):
        """str: *READ ONLY* Identifies the specific image projection type supplied."""
        for attribute in self._choice[0]['collection']:
            if getattr(self, attribute) is not None:
                return attribute
        return None

    @property
    def ReferencePoint(self):
        """
        None|ReferencePointType: *READ ONLY* Gets the reference point.
        """

        if self.ProjectionType:
            return getattr(self, self.ProjectionType).ReferencePoint
        return None



"""
Simple helper functions for constructing the NGA modified AFRL/RDE structure 
assuming either a known ground truth scenario or inferred analyst truth 
scenario.
"""

__classification__ = 'UNCLASSIFIED'
__author__ = "Thomas McCullough"

from typing import List, Dict, Union, Optional

import numpy

from sarpy.io.complex.sicd_elements.SICD import SICDType
from sarpy.io.complex.sicd import SICDReader

from sarpy.annotation.afrl_rde_elements.blocks import LabelSourceType
from sarpy.annotation.afrl_rde_elements.Research import ResearchType
from sarpy.annotation.afrl_rde_elements.CollectionInfo import CollectionInfoType
from sarpy.annotation.afrl_rde_elements.SubCollectionInfo import SubCollectionInfoType
from sarpy.annotation.afrl_rde_elements.ObjectInfo import ObjectInfoType, \
    TheObjectType, GeoLocationType as ObjectGeoLocation, \
    ImageLocationType as ObjectImageLocation, SizeType, OrientationType, \
    StringWithComponentType
from sarpy.annotation.afrl_rde_elements.FiducialInfo import FiducialInfoType, \
    TheFiducialType, GeoLocationType as FiducialGeoLocation, \
    ImageLocationType as FiducialImageLocation
from sarpy.annotation.afrl_rde_elements.ImageInfo import ImageInfoType
from sarpy.annotation.afrl_rde_elements.SensorInfo import SensorInfoType

from sarpy.annotation.label import LabelSchema, FileLabelCollection, LabelCollection, \
    LabelFeature, LabelProperties, LabelMetadata


class GroundTruthConstructor(object):
    """
    This class is a helper for performing a ground truth construction.
    """

    __slots__ = (
        '_collection_info', '_subcollection_info', '_label_source', '_objects', '_fiducials')

    def __init__(
            self, collection_info: CollectionInfoType,
            subcollection_info: SubCollectionInfoType,
            label_source: Optional[LabelSourceType] = None):
        """

        Parameters
        ----------
        collection_info : CollectionInfoType
        subcollection_info : SubCollectionInfoType
        label_source : None|LabelSourceType
        """

        self._collection_info = collection_info
        self._subcollection_info = subcollection_info
        if label_source is None:
            self._label_source = LabelSourceType(SourceType='Ground Truth', SourceID='Unspecified')
        else:
            self._label_source = label_source
        self._objects = []
        self._fiducials = []

    def add_fiducial(self, the_fiducial: TheFiducialType) -> None:
        """
        Adds the given fiducial to the collection.

        Parameters
        ----------
        the_fiducial : TheFiducialType
        """

        if not isinstance(the_fiducial, TheFiducialType):
            raise TypeError('Requires an object of type `TheFiducialType`, got `{}`'.format(type(the_fiducial)))
        if the_fiducial.ImageLocation is not None:
            raise ValueError('The fiducial has ImageLocation already set.')
        if the_fiducial.SlantPlane is not None or the_fiducial.GroundPlane is not None:
            raise ValueError('The fiducial already has the SlantPlane or GroundPlane set.')
        self._fiducials.append(the_fiducial)

    def add_fiducial_from_arguments(
            self,
            Name: str = None,
            SerialNumber: Optional[str] = None,
            FiducialType: Optional[str] = None,
            GeoLocation: FiducialGeoLocation = None) -> None:
        """
        Adds a fiducial to the collection.

        Parameters
        ----------
        Name : str
        SerialNumber : None|str
        FiducialType : None|str
        GeoLocation : FiducialGeoLocation
        """

        self.add_fiducial(
            TheFiducialType(
                Name=Name,
                SerialNumber=SerialNumber,
                FiducialType=FiducialType,
                GeoLocation=GeoLocation))

    def add_object(
            self,
            the_object: TheObjectType) -> None:
        """
        Adds the given object to the collection.

        Parameters
        ----------
        the_object : TheObjectType
        """

        if not isinstance(the_object, TheObjectType):
            raise TypeError('Requires an object of type `TheObjectType`, got `{}`'.format(type(the_object)))
        if the_object.ImageLocation is not None:
            raise ValueError('The object has ImageLocation already set.')
        if the_object.SlantPlane is not None or the_object.GroundPlane is not None:
            raise ValueError('The object already has the SlantPlane or GroundPlane set.')
        self._objects.append(the_object)

    def add_object_from_arguments(
            self,
            SystemName: str = None,
            SystemComponent: Optional[str] = None,
            NATOName: Optional[str] = None,
            Function: Optional[str] = None,
            Version: Optional[str] = None,
            DecoyType: Optional[str] = None,
            SerialNumber: Optional[str] = None,
            ObjectClass: str = 'Unknown',
            ObjectSubClass: str = 'Unknown',
            ObjectTypeClass: str = 'Unknown',
            ObjectType: str = 'Unknown',
            ObjectLabel: str = None,
            Size: Optional[Union[SizeType, numpy.ndarray, list, tuple]] = None,
            Orientation: OrientationType = None,
            Articulation: Union[None, str, StringWithComponentType, List[StringWithComponentType]] = None,
            Configuration: Union[None, str, StringWithComponentType, List[StringWithComponentType]] = None,
            Accessories: Optional[str] = None,
            PaintScheme: Optional[str] = None,
            Camouflage: Optional[str] = None,
            Obscuration: Optional[str] = None,
            ObscurationPercent: Optional[float] = None,
            ImageLevelObscuration: Optional[str] = None,
            GeoLocation: ObjectGeoLocation = None,
            TargetToClutterRatio: Optional[str] = None,
            VisualQualityMetric: Optional[str] = None,
            UnderlyingTerrain: Optional[str] = None,
            OverlyingTerrain: Optional[str] = None,
            TerrainTexture: Optional[str] = None,
            SeasonalCover: Optional[str] = None) -> None:
        """
        Adds an object to the collection.

        Parameters
        ----------
        SystemName : str
        SystemComponent : None|str
        NATOName : None|str
        Function : None|str
        Version : None|str
        DecoyType : None|str
        SerialNumber : None|str
        ObjectClass : None|str
        ObjectSubClass : None|str
        ObjectTypeClass : None|str
        ObjectType : None|str
        ObjectLabel : None|str
        Size : None|SizeType|numpy.ndarray|list|tuple
        Orientation : OrientationType
        Articulation : None|str|StringWithCompoundType|List[StringWithCompoundType]
        Configuration : None|str|StringWithCompoundType|List[StringWithCompoundType]
        Accessories : None|str
        PaintScheme : None|str
        Camouflage : None|str
        Obscuration : None|str
        ObscurationPercent : None|float
        ImageLevelObscuration : None|str
        GeoLocation : ObjectGeoLocation
        TargetToClutterRatio : None|str
        VisualQualityMetric : None|str
        UnderlyingTerrain : None|str
        OverlyingTerrain : None|str
        TerrainTexture : None|str
        SeasonalCover : None|str
        """

        self.add_object(
            TheObjectType(SystemName=SystemName,
                          SystemComponent=SystemComponent,
                          NATOName=NATOName,
                          Function=Function,
                          Version=Version,
                          DecoyType=DecoyType,
                          SerialNumber=SerialNumber,
                          ObjectClass=ObjectClass,
                          ObjectSubClass=ObjectSubClass,
                          ObjectTypeClass=ObjectTypeClass,
                          ObjectType=ObjectType,
                          ObjectLabel=ObjectLabel,
                          Size=Size,
                          Orientation=Orientation,
                          Articulation=Articulation,
                          Configuration=Configuration,
                          Accessories=Accessories,
                          PaintScheme=PaintScheme,
                          Camouflage=Camouflage,
                          Obscuration=Obscuration,
                          ObscurationPercent=ObscurationPercent,
                          ImageLevelObscuration=ImageLevelObscuration,
                          GeoLocation=GeoLocation,
                          TargetToClutterRatio=TargetToClutterRatio,
                          VisualQualityMetric=VisualQualityMetric,
                          UnderlyingTerrain=UnderlyingTerrain,
                          OverlyingTerrain=OverlyingTerrain,
                          TerrainTexture=TerrainTexture,
                          SeasonalCover=SeasonalCover))

    def get_final_structure(self) -> ResearchType:
        """
        It is anticipated that this might be reused to localize for a whole series
        of different sicd files.

        Gets **a static copy** of the constructed AFRL Research structure. This has the
        provided CollectionInfo and SubCollectionInfo populated. It also
        has the ObjectInfo and FiducialInfo with the GeoLocation
        ground truth details that have been provided.

        No image location information has been populated, and there are no
        ImageInfo or SensorInfo populated, because these are independent
        of ground truth.

        Returns
        -------
        ResearchType
        """

        return ResearchType(
            DetailCollectionInfo=self._collection_info,
            DetailSubCollectionInfo=self._subcollection_info,
            DetailFiducialInfo=FiducialInfoType(
                NumberOfFiducialsInScene=len(self._fiducials),
                NumberOfFiducialsInImage=len(self._fiducials),
                LabelSource=self._label_source,
                Fiducials=self._fiducials),
            DetailObjectInfo=ObjectInfoType(
                NumberOfObjectsInScene=len(self._objects),
                NumberOfObjectsInImage=len(self._objects),
                LabelSource=self._label_source,
                Objects=self._objects)).copy()

    def localize_for_sicd(
            self,
            sicd: SICDType,
            base_sicd_file: str,
            layover_shift: bool = False,
            populate_in_periphery: bool = False,
            include_out_of_range: bool = False,
            padding_fraction: Optional[float] = 0.05,
            minimum_pad: Union[int, float] = 0,
            md5_checksum: Optional[str] = None):
        """
        Localize the AFRL structure for the given sicd structure.

        This returns **a static copy** of the AFRL structure, and this method
        can be repeatedly applied for a sequence of different sicd files which all
        apply to the same ground truth scenario.

        Parameters
        ----------
        sicd : SICDType
        base_sicd_file : str
        layover_shift : bool
        populate_in_periphery : bool
        include_out_of_range : bool
        padding_fraction : None|float
        minimum_pad : int|float
        md5_checksum : None|str

        Returns
        -------
        ResearchType
        """

        out_research = self.get_final_structure()
        out_research.apply_sicd(
            sicd,
            base_sicd_file,
            layover_shift=layover_shift,
            populate_in_periphery=populate_in_periphery,
            include_out_of_range=include_out_of_range,
            padding_fraction=padding_fraction,
            minimum_pad=minimum_pad,
            md5_checksum=md5_checksum)
        return out_research

    def localize_for_sicd_reader(
            self,
            sicd_reader: SICDReader,
            layover_shift: bool = False,
            populate_in_periphery: bool = False,
            include_out_of_range: bool = False,
            padding_fraction: Optional[float] = 0.05,
            minimum_pad: Union[int, float] = 0,
            populate_md5: bool = True):
        """
        Localize the AFRL structure for the given sicd file.

        This returns **a static copy** of the AFRL structure, and this method
        can be repeatedly applied for a sequence of different sicd files which all
        apply to the same ground truth scenario.

        Parameters
        ----------
        sicd_reader : SICDReader
        layover_shift : bool
        populate_in_periphery : bool
        include_out_of_range : bool
        padding_fraction : None|float
        minimum_pad : int|float
        populate_md5 : bool

        Returns
        -------
        ResearchType
        """

        out_research = self.get_final_structure()
        out_research.apply_sicd_reader(
            sicd_reader,
            layover_shift=layover_shift,
            populate_in_periphery=populate_in_periphery,
            include_out_of_range=include_out_of_range,
            padding_fraction=padding_fraction,
            minimum_pad=minimum_pad,
            populate_md5=populate_md5)
        return out_research


class AnalystTruthConstructor(object):
    """
    This class is a helper for performing an analyst truth construction.
    """

    __slots__ = (
        '_sicd', '_base_file',
        '_collection_info', '_subcollection_info', '_image_info', '_sensor_info',
        '_label_source', '_objects', '_fiducials',
        '_projection_type', '_proj_kwargs')

    def __init__(
            self,
            sicd: SICDType,
            base_file: str,
            collection_info: CollectionInfoType,
            subcollection_info: SubCollectionInfoType,
            label_source: Optional[LabelSourceType] = None,
            projection_type: str = 'HAE',
            proj_kwargs: Optional[Dict] = None,
            md5_checksum: Optional[str] = None):
        """

        Parameters
        ----------
        sicd : SICDType
        base_file : str
        collection_info : CollectionInfoType
        subcollection_info : SubCollectionInfoType
        label_source : None|LabelSourceType
        projection_type : str
            One of 'PLANE', 'HAE', or 'DEM'. The value of `proj_kwargs`
            will need to be appropriate.
        proj_kwargs : None|Dict
            The keyword arguments for the :func:`SICDType.project_image_to_ground_geo` method.
        md5_checksum : None|str
            The MD5 checksum of the full image file.
        """

        self._sicd = sicd
        self._base_file = base_file

        # TODO: should we create a decent shell for general Analyst Truth
        #  collection and subcollection info?
        self._collection_info = collection_info
        self._subcollection_info = subcollection_info
        self._image_info = ImageInfoType.from_sicd(self._sicd, self._base_file, md5_checksum=md5_checksum)
        self._sensor_info = SensorInfoType.from_sicd(self._sicd)
        if label_source is None:
            self._label_source = LabelSourceType(SourceType='Analyst Truth', SourceID='Unspecified')
        else:
            self._label_source = label_source

        self._objects = []
        self._fiducials = []

        self._projection_type = projection_type
        self._proj_kwargs = {} if proj_kwargs is None else proj_kwargs

    @property
    def image_info(self) -> ImageInfoType:
        """
        ImageInfoType: The basic image info object derived from the sicd
        """

        return self._image_info

    @property
    def sensor_info(self) -> SensorInfoType:
        """
        SensorInfoType: The basic sensor info object derived from the sicd.
        """

        return self._sensor_info

    def add_fiducial(self, the_fiducial: TheFiducialType) -> None:
        """
        Adds the given fiducial to the collection. Note that this object will be modified in place.

        Parameters
        ----------
        the_fiducial : TheFiducialType
        """

        if not isinstance(the_fiducial, TheFiducialType):
            raise TypeError('Requires an object of type `TheFiducialType`, got `{}`'.format(type(the_fiducial)))
        if the_fiducial.GeoLocation is not None:
            raise ValueError('The fiducial has GeoLocation already set.')
        the_fiducial.set_geo_location_from_sicd(self._sicd, projection_type=self._projection_type, **self._proj_kwargs)
        self._fiducials.append(the_fiducial)

    def add_fiducial_from_arguments(
            self,
            Name: Optional[str] = None,
            SerialNumber: Optional[str] = None,
            FiducialType: Optional[str] = None,
            ImageLocation: FiducialImageLocation = None):
        """
        Adds a fiducial to the collection.

        Parameters
        ----------
        Name : None|str
        SerialNumber : None|str
        FiducialType : None|str
        ImageLocation : FiducialImageLocation
        """

        self.add_fiducial(
            TheFiducialType(
                Name=Name,
                SerialNumber=SerialNumber,
                FiducialType=FiducialType,
                ImageLocation=ImageLocation))

    def add_object(
            self,
            the_object: TheObjectType,
            padding_fraction: Optional[float] = 0.05,
            minimum_pad: Union[int, float] = 0):
        """
        Adds the object to the collection. Note that this object will be modified in place.

        Parameters
        ----------
        the_object : TheObjectType
        padding_fraction : None|float
            Default fraction of box dimension by which to pad.
        minimum_pad : float|int
            The minimum number of pixels by which to pad for the chip
        """

        if not isinstance(the_object, TheObjectType):
            raise TypeError('Requires an object of type `TheObjectType`, got `{}`'.format(type(the_object)))
        if the_object.GeoLocation is not None:
            raise ValueError('The object has GeoLocation already set.')
        the_object.set_geo_location_from_sicd(
            self._sicd, projection_type=self._projection_type, **self._proj_kwargs)
        the_object.set_chip_details_from_sicd(
            self._sicd, populate_in_periphery=True, padding_fraction=padding_fraction, minimum_pad=minimum_pad)
        self._objects.append(the_object)

    def add_object_from_arguments(
            self,
            padding_fraction: float = 0.05,
            minimum_pad: Union[int, float] = 0,
            SystemName: str = None,
            SystemComponent: Optional[str] = None,
            NATOName: Optional[str] = None,
            Function: Optional[str] = None,
            Version: Optional[str] = None,
            DecoyType: Optional[str] = None,
            SerialNumber: Optional[str] = None,
            ObjectClass: str = 'Unknown',
            ObjectSubClass: str = 'Unknown',
            ObjectTypeClass: str = 'Unknown',
            ObjectType: str = 'Unknown',
            ObjectLabel: str = None,
            Size: Union[None, SizeType, numpy.ndarray, list, tuple] = None,
            Orientation: OrientationType = None,
            Articulation: Union[None, str, StringWithComponentType, List[StringWithComponentType]] = None,
            Configuration: Union[None, str, StringWithComponentType, List[StringWithComponentType]] = None,
            Accessories: Optional[str] = None,
            PaintScheme: Optional[str] = None,
            Camouflage: Optional[str] = None,
            Obscuration: Optional[str] = None,
            ObscurationPercent: Optional[float] = None,
            ImageLevelObscuration: Optional[str] = None,
            ImageLocation: ObjectImageLocation = None,
            TargetToClutterRatio: Optional[str] = None,
            VisualQualityMetric: Optional[str] = None,
            UnderlyingTerrain: Optional[str] = None,
            OverlyingTerrain: Optional[str] = None,
            TerrainTexture: Optional[str] = None,
            SeasonalCover: Optional[str] = None) -> None:
        """
        Adds an object to the collection.

        Parameters
        ----------
        padding_fraction : None|float
            Default fraction of box dimension by which to pad.
        minimum_pad : float|int
        SystemName : str
        SystemComponent : None|str
        NATOName : None|str
        Function : None|str
        Version : None|str
        DecoyType : None|str
        SerialNumber : None|str
        ObjectClass : None|str
        ObjectSubClass : None|str
        ObjectTypeClass : None|str
        ObjectType : None|str
        ObjectLabel : None|str
        Size : None|SizeType|numpy.ndarray|list|tuple
        Orientation : OrientationType
        Articulation : None|str|StringWithComponentType|List[StringWithComponentType]
        Configuration : None|str|StringWithComponentType|List[StringWithComponentType]
        Accessories : None|str
        PaintScheme : None|str
        Camouflage : None|str
        Obscuration : None|str
        ObscurationPercent : None|float
        ImageLevelObscuration : None|str
        ImageLocation : ObjectImageLocation
        TargetToClutterRatio : None|str
        VisualQualityMetric : None|str
        UnderlyingTerrain : None|str
        OverlyingTerrain : None|str
        TerrainTexture : None|str
        SeasonalCover : None|str
        """

        self.add_object(
            TheObjectType(SystemName=SystemName,
                          SystemComponent=SystemComponent,
                          NATOName=NATOName,
                          Function=Function,
                          Version=Version,
                          DecoyType=DecoyType,
                          SerialNumber=SerialNumber,
                          ObjectClass=ObjectClass,
                          ObjectSubClass=ObjectSubClass,
                          ObjectTypeClass=ObjectTypeClass,
                          ObjectType=ObjectType,
                          ObjectLabel=ObjectLabel,
                          Size=Size,
                          Orientation=Orientation,
                          Articulation=Articulation,
                          Configuration=Configuration,
                          Accessories=Accessories,
                          PaintScheme=PaintScheme,
                          Camouflage=Camouflage,
                          Obscuration=Obscuration,
                          ObscurationPercent=ObscurationPercent,
                          ImageLevelObscuration=ImageLevelObscuration,
                          ImageLocation=ImageLocation,
                          TargetToClutterRatio=TargetToClutterRatio,
                          VisualQualityMetric=VisualQualityMetric,
                          UnderlyingTerrain=UnderlyingTerrain,
                          OverlyingTerrain=OverlyingTerrain,
                          TerrainTexture=TerrainTexture,
                          SeasonalCover=SeasonalCover),
            padding_fraction=padding_fraction,
            minimum_pad=minimum_pad)

    def get_final_structure(self) -> ResearchType:
        """
        This is not anticipated to be reused, so the raw progress to date is returned.
        Care should be taken in modifying the returned structure directly.

        Returns
        -------
        ResearchType
        """

        return ResearchType(
            DetailCollectionInfo=self._collection_info,
            DetailSubCollectionInfo=self._subcollection_info,
            DetailImageInfo=self._image_info,
            DetailSensorInfo=self._sensor_info,
            DetailFiducialInfo=FiducialInfoType(
                NumberOfFiducialsInScene=len(self._fiducials),
                NumberOfFiducialsInImage=len(self._fiducials),
                LabelSource=self._label_source,
                Fiducials=self._fiducials),
            DetailObjectInfo=ObjectInfoType(
                NumberOfObjectsInScene=len(self._objects),
                NumberOfObjectsInImage=len(self._objects),
                LabelSource=self._label_source,
                Objects=self._objects))


def convert_afrl_to_native(
        research: ResearchType,
        include_chip: bool = False) -> FileLabelCollection:
    """
    Converts an AFRL structure to a label structure for simple viewing.

    Parameters
    ----------
    research : ResearchType
    include_chip : bool
        Include the chip definition in the geometry structure?

    Returns
    -------
    FileLabelCollection
    """

    def _convert_object_to_json(t_object: TheObjectType) -> LabelFeature:
        # extract the "properties"
        geometry, geometry_properties = t_object.get_image_geometry_object_for_sicd(include_chip=include_chip)
        feature = LabelFeature(
            geometry=geometry,
            properties=LabelProperties(
                name=t_object.SystemName,
                geometry_properties=geometry_properties))
        feature.add_annotation_metadata(LabelMetadata(label_id=t_object.ObjectLabel))
        return feature

    if not isinstance(research, ResearchType):
        raise TypeError('Expected ResearchType, got type `{}`'.format(type(research)))
    if research.DetailObjectInfo is None or \
            research.DetailObjectInfo.Objects is None or \
            len(research.DetailObjectInfo.Objects) == 0:
        raise ValueError('Nothing to be done')

    # create our adhoc containers
    label_schema = LabelSchema(version='AdHoc')
    annotation_collection = LabelCollection()
    for the_object in research.DetailObjectInfo.Objects:
        new_key = the_object.ObjectLabel
        if new_key not in label_schema.labels:
            label_schema.add_entry(new_key, new_key)
        annotation_collection.add_feature(_convert_object_to_json(the_object))

    # finalize the collection
    return FileLabelCollection(
        label_schema,
        annotations=annotation_collection,
        image_file_name=research.DetailImageInfo.DataFilename)



"""
Definition for the CollectionInfo NGA modified RDE/AFRL labeling object
"""

__classification__ = "UNCLASSIFIED"
__authors__ = "Thomas McCullough"


from typing import Optional

from sarpy.io.xml.base import Serializable
from sarpy.io.xml.descriptors import IntegerDescriptor, SerializableDescriptor, \
    StringDescriptor, FloatDescriptor

from .base import DEFAULT_STRICT
from .blocks import DateTimeRangeType


class LocationType(Serializable):
    _fields = ('Lat', 'Lon', 'Name')
    _required = ('Lat', 'Lon')
    # descriptors
    Lat = FloatDescriptor(
        'Lat', _required, strict=DEFAULT_STRICT,
        docstring="General latitude of the data collection.")  # type: float
    Lon = FloatDescriptor(
        'Lon', _required, strict=DEFAULT_STRICT,
        docstring="General longitude of the data collection.")  # type: float
    Name = StringDescriptor(
        'Name', _required,
        docstring="Common name of the collection location.")  # type: Optional[str]

    def __init__(self, Lat=None, Lon=None, Name=None, **kwargs):
        """
        Parameters
        ----------
        Lat : float
        Lon : float
        Name : None|str
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Lat = Lat
        self.Lon = Lon
        self.Name = Name
        super(LocationType, self).__init__(**kwargs)


class CollectionInfoType(Serializable):
    _fields = (
        'Name', 'ProgramName', 'Sponsor', 'Date', 'Location', 'NumberOfSites')
    _required = ('Date', )
    # descriptors
    Name = StringDescriptor(
        'Name', _required,
        docstring="Name of the collection.")  # type: Optional[str]
    ProgramName = StringDescriptor(
        'StringDescriptor', _required,
        docstring="Name of the program that collected the data.")  # type: Optional[str]
    Sponsor = StringDescriptor(
        'Sponsor', _required,
        docstring="Sponsoring agency/organization of the data collection.")  # type: Optional[str]
    Date = SerializableDescriptor(
        'Date', DateTimeRangeType, _required, strict=DEFAULT_STRICT,
        docstring="Begin and end dates of the data collection.")  # type: Optional[DateTimeRangeType]
    Location = SerializableDescriptor(
        'Location', LocationType, _required, strict=DEFAULT_STRICT,
        docstring="General location of the data collection.")  # type: Optional[LocationType]
    NumberOfSites = IntegerDescriptor(
        'NumberOfSites', _required, strict=DEFAULT_STRICT,
        docstring="Number of different sites contained in the data collection.")  # type: Optional[int]

    def __init__(self, Name=None, ProgramName=None, Sponsor=None, Date=None,
                 Location=None, NumberOfSites=None, **kwargs):
        """
        Parameters
        ----------
        Name : None|str
        ProgramName : None|str
        Sponsor : None|str
        Date : None|DateTimeRangeType|list|tuple
        Location : None|LocationType
        NumberOfSites : None|int
        kwargs
            Other keyword arguments
        """

        if '_xml_ns' in kwargs:
            self._xml_ns = kwargs['_xml_ns']
        if '_xml_ns_key' in kwargs:
            self._xml_ns_key = kwargs['_xml_ns_key']
        self.Name = Name
        self.ProgramName = ProgramName
        self.Sponsor = Sponsor
        self.Date = Date
        self.Location = Location
        self.NumberOfSites = NumberOfSites
        super(CollectionInfoType, self).__init__(**kwargs)
